import argparse
from types import SimpleNamespace
from pathlib import Path
from datetime import datetime
import sys

import rtoml

from boss.utils import init_logger


"""
Configuration:
1. load a template toml that contains defaults
2. parse a toml given on command line
3. overwrite defaults with CL toml
4. parse a toml for readfish separately
5. exchange args between the two
"""



class Config:
    def __init__(self, parse: bool = False, arg_list: list = None):
        """
        Initialise configuration by loading defaults
        When debugging, paths to the TOMLs can be given as a list or arguments to parse
        For simulations, providing a readfish TOML is not necessary

        :param parse: Parse from command line/toml_paths
        :param arg_list: Pass-through list of paths to toml files
        """
        self.template_toml = """
        [general]
        name = "boss"                   # experiment name
        wait = 60                       # waiting time between updates in live version
        ref = ""                        # reference file (fasta or None). Not specifying a file switches to AEONS
        mmi = ""                        # index of reference (will be built if ref is given but not mmi)
        
        [live]
        device = "TEST"                 # position on sequencer
        host = "localhost"              # host of sequencing device
        port = 9502                     # port of sequencing device
        data_wait = 100                 # wait for X Mb of data before first update

        [optional]
        reject_refs = ""                # comma-separated list of headers in reference from which to always reject
        ploidy = 1                      # 1 or 2
        lowcov = 10                     # target coverage for assemblies
        temperature = 60                # max updates during which to consider fragments (higher number might decrease update speeds)
        min_seq_len = 2500              # min sequence length used during contig reconstruction
        min_contig_len = 10_000         # min length to consider contigs
        min_s1 = 200                    # min alignment scores to consider overlaps
        min_map_len = 2000              # min alignment length to consider overlaps
        tetra = true                    # perform tetranucleotide frequency tests
        filter_repeats = false          # perform repeat filtering
        bucket_threshold = 5            # at which coverage to switch on the strategy in a bucket (debug)

        ########################################################################################

        [simulation]                    # simulation arguments
        fq = ""
        batchsize = 4000                    
        maxb = 400
        binit = 5
        dumptime = 200000000
        paf_full = ""                   # giving pafs triggers BOSS-RUNS
        paf_trunc = ""
        """
        # load the default from above
        self.args = rtoml.loads(self.template_toml)
        # convert toml dict to SimpleNamespace
        self.args = self._convert_to_namespace()
        # do we parse toml paths to overwrite defaults?
        if parse:
            self.arg_list = arg_list
            # load TOMLs from command line if not given as list
            if not self.arg_list:
                self.arg_list = sys.argv[1:]

            toml_paths = self._parse_toml_args()
            args_file, args_readfish = self._load_tomls(toml_paths)
            self._overwrite_defaults(args_file)
            # check if we are simulating or running real experiment
            self._check_run_type()
            # initialise a log file in the output folder
            stamp = datetime.now().strftime("%Y%m%d-%H%M%S")
            self.logfile = f'{stamp}_boss.log'
            init_logger(logfile=self.logfile, args=self.args)
            # config settings for readfish
            if self.args.live_run:
                # add path to readfish toml as arg
                self.args.toml_readfish = toml_paths.toml_readfish
                # check that experiment exists as readfish region
                self._verify_region_names(self.args, args_readfish)
                # validate readfish args
                self._validate_readfish_conf(args_readfish)



    def _check_run_type(self) -> None:
        """
        Check if we are running a simulation or a real experiment
        If "fq" under the simulation header is given, we simulate

        :return:
        """
        if self.args.fq:
            self.args.sim_run = True
            self.args.live_run = False
        elif self.args.device:
            self.args.live_run = True
            self.args.sim_run = False
        else:
            raise ValueError("Need either fastq for simulation or device for live run")



    def _convert_to_namespace(self) -> SimpleNamespace:
        """
        Convert arguments from a parsed Dict to a Namespace object
        For method-style access to attributes

        :return: Arguments as SimpleNamespace object
        """
        args = SimpleNamespace()
        for category, subdict in self.args.items():
            if not type(subdict) is dict:
                setattr(args, category, subdict)
            else:
                for k, v in subdict.items():
                    setattr(args, k, v)
        return args



    def _parse_toml_args(self) -> argparse.Namespace:
        """
        Parse TOML paths given as arguments on the command line

        :return: argparse.Namespace with paths to the two TOMLs
        """
        parser = argparse.ArgumentParser()
        parser.add_argument(
            '--toml',
            type=str,
            default=None,
            required=True,
            help='TOML configuration file')
        parser.add_argument(
            '--toml_readfish',
            type=str,
            default=None,
            help='TOML configuration file for readfish')
        toml_paths = parser.parse_args(self.arg_list)
        return toml_paths



    @staticmethod
    def _load_tomls(toml_paths: argparse.Namespace) -> tuple[dict, dict]:
        """
        Load the TOML files into dictionaries using rTOML

        :param toml_paths: Paths to TOMLs as parsed arguments
        :return: Tuple with parsed TOML dictionaries
        """
        args_file = rtoml.loads(Path(toml_paths.toml).read_text(encoding="utf-8"))
        if toml_paths.toml_readfish:
            args_readfish_file = rtoml.loads(Path(toml_paths.toml_readfish).read_text(encoding="utf-8"))
        else:
            args_readfish_file = dict()
        return args_file, args_readfish_file



    def _overwrite_defaults(self, args_from_file: dict) -> None:
        """
        Use TOML given on CL to overwrite defaults

        :param args_from_file: parsed contents of given TOML
        :return:
        """
        for category in args_from_file:
            for k, v in args_from_file[category].items():
                setattr(self.args, k, v)



    @staticmethod
    def _verify_region_names(args: SimpleNamespace, args_readfish: dict) -> None:
        """
        Verify that the experiment name of BOSS exists as region in readfish

        :param args: Config dictionary for BOSS
        :param args_readfish: Config dictionary for readfish
        :return:
        """
        if type(args_readfish['regions']) is not list:
            raise ValueError("Readfish regions must be specified as array")

        # make sure the names of BOSS and regions on flowcell are the same
        region_names = {r['name'] for r in args_readfish['regions']}
        if args.name not in region_names:
            raise ValueError("One of the regions in readfish needs the same name as the experiment in BOSS")



    @staticmethod
    def _validate_readfish_conf(args_rf: dict) -> int:
        """
        Minimalist version of readfish validate entry-point
        to validate parsed toml config for readfish

        :param args_rf: Dict of arguments parsed for readfish
        :return: indicator
        """
        from readfish._config import Conf
        channels = 512
        try:
            _ = Conf.from_dict(args_rf, channels)
        except:
            raise ValueError("Could not load TOML config for readfish")
        return 0








import logging
import os
from collections import defaultdict, Counter
from copy import deepcopy
from shutil import copy
import time
from types import SimpleNamespace
from typing import Optional, Any
from concurrent.futures import ThreadPoolExecutor as TPexe
from pathlib import Path

import numpy as np
from numpy.typing import NDArray
import bottleneck as bn

from boss.utils import execute, random_id, load_gfa, write_logs
from boss.dependencies import Dependencies
from boss.paf import Paf, PafLine
from boss.aeons.kmer import euclidean_dist, euclidean_threshold


Edge = tuple[str, str]  # typehint for containments: tuple of (source, target)



class SequenceAVA:

    def __init__(self, paf: str, filters: SimpleNamespace, tetra: bool = False):
        """
        Initialize the SequenceAVA object.

        :param paf: Path to the PAF file.
        :param filters: filters to apply, stored as arguments.
        :param tetra: Whether to use the tetramer distance to filter overlaps. Defaults to False.
        """
        self.paf = paf
        self.gfa = f'{paf}.gfa'
        self.filters = filters
        self.tetra = tetra
        # container to keep overlaps
        # save paf lines as qname-tname with alphanum sort
        self.links = defaultdict(lambda: defaultdict(PafLine))
        self.paf_links = f"{paf}.links.paf"
        self.dep = Dependencies()


    def load_ava(self, paf: str, seqpool: "SequencePool") -> tuple[dict[Edge, PafLine], set]:
        """
        Load all entries from a PAF file as PafLines and filter the entries while loading.

        :param paf: Path to the PAF file.
        :param seqpool: SequencePool
        :return: A tuple containing containments and overlappers.
        """
        self.trims = []  # used for trimming
        self.overlaps = {}  # used for trimming
        containments = {}  # collect, used for coverage increments
        overlappers = set()  # used to track temperature of sequences
        ovl = 0
        inter = 0

        records, skip = Paf.parse_filter_classify_records(paf=paf, filters=self.filters)

        for rec in records:
            if rec.c == 2:
                # first contained
                # check if we have a containment of the two already
                if (rec.qname, rec.tname) in containments.keys():
                    if rec.s1 < containments[(rec.qname, rec.tname)].s1:
                        # previously recorded containment has higher s1
                        continue
                containments[(rec.qname, rec.tname)] = rec
            elif rec.c == 3:
                # second contained
                # check if we have a containment of the two already
                if (rec.tname, rec.qname) in containments.keys():
                    if rec.s1 < containments[(rec.tname, rec.qname)].s1:
                        # previously recorded containment has higher s1
                        continue
                containments[(rec.tname, rec.qname)] = rec
            elif rec.c in {4, 5}:
                # if applicable, check for tetramer dist
                if self.tetra:
                    intra = seqpool.is_intra(rec.qname, rec.tname)
                    if not intra:
                        inter += 1
                        continue

                if not seqpool.sequences[rec.tname].acceptor:
                    rec.c = 2
                    containments[(rec.qname, rec.tname)] = rec
                    continue
                if not seqpool.sequences[rec.qname].acceptor:
                    rec.c = 3
                    containments[(rec.tname, rec.qname)] = rec
                    continue

                ovl += 1
                self.overlaps[(rec.qname, rec.tname)] = rec
                # check if we have an overlap of the two already
                if rec.tname in self.links[rec.qname].keys():
                    if rec.s1 < self.links[rec.qname][rec.tname].s1:
                        # previously recorded overlap has higher s1
                        continue

                self.links[rec.qname][rec.tname] = rec
                self.links[rec.tname][rec.qname] = rec
                overlappers.add(rec.qname)
                overlappers.add(rec.tname)
            elif rec.c == 6:
                self.trims.append(rec)
            else:
                pass

        contained_ids = set([i for (i, j) in containments.keys()])
        # remove contained ones from further containment search
        skip_filt = [s for s in skip if s.qname not in contained_ids and s.tname not in contained_ids]
        # run multiline containment detection
        mc = MultilineContainments(records=skip_filt)
        containments.update(mc.containments)

        logging.info(f"ava load: skip {len(skip)}, cont {len(contained_ids)} cont multi: {len(mc.containments)} ovl: {ovl} inter: {inter}")
        return containments, overlappers


    def remove_links(self, sequences: set[str]) -> None:
        """
        Remove overlaps of certain sequences.

        :param sequences: Set of sequence IDs to remove overlaps from.
        """
        for sid in sequences:
            # targets for overlaps where sid is "query"
            targets = self.links[sid].keys()
            # remove the overlaps where sid is "query"
            self.links.pop(sid, None)
            # remove overlaps from targets
            for t in targets:
                self.links[t].pop(sid, None)


    def to_be_trimmed(self) -> dict[str, tuple[int, int, Any]]:
        """
        After classification, find the coordinates for trimming.

        :return: A dictionary containing sequences to be trimmed and their corresponding coordinates.
        """
        trim = self.trims
        # save the name and coordinates for trimming
        to_trim = {}
        for rec in trim:
            sid, trim_start, trim_stop, other = rec.find_trim_coords()
            if sid == '0':  # don't trim if product would be short
                continue
            to_trim[sid] = (trim_start, trim_stop, other)
        return to_trim


    @staticmethod
    def trim_success(trim_dict: dict[str, tuple[int, int, Any]], overlaps: dict[tuple[str, str], Any]) -> set[str]:
        """
        Check which trimming attempts were successful.

        :param trim_dict: Dictionary containing sequences to be trimmed and their coordinates.
        :param overlaps: Dictionary of overlaps.
        :return: A set of sequences to be removed.
        """
        success, unsuccess = set(), set()
        trim = set(trim_dict.keys())
        if not trim:
            to_remove = success | unsuccess
            return to_remove
        if not overlaps:
            unsuccess = {f'{t}%' for t in trim}
            to_remove = success | unsuccess
            return to_remove

        ovl_q, ovl_t = zip(*overlaps.keys())
        ovl = set(ovl_q) | set(ovl_t)
        trim_mod = {f'{t}%' for t in trim}
        success_raw = trim_mod & ovl
        unsuccess = trim_mod - success_raw
        # remove the percentage marker for removal
        success = {s[:-1] for s in success_raw}
        # merge the headers for removal
        to_remove = success | unsuccess
        return to_remove


    def links2paf(self, paf_out: str) -> None:
        """
        Write overlaps to a PAF file.

        :param paf_out: Output file path for the PAF file.
        """
        written = set()
        with open(paf_out, 'w') as fh:
            for node, target_dict in self.links.items():
                if not target_dict:
                    continue
                # go through all links
                for target, rec in target_dict.items():
                    if rec.line in written:
                        continue
                    else:
                        fh.write(rec.line)
                        written.add(rec.line)


    def paf2gfa_gfatools(self, paf: str, fa: str, gfa: Optional[str] = None) -> str:
        """
        Use gfatools to generate the graph and unitigs from a PAF file and a FASTA file.

        :param paf: Input PAF file path.
        :param fa: Input FASTA file path.
        :param gfa: Optional output GFA file path.
        :return: The generated GFA content as a string or empty string
        """
        if not os.path.getsize(paf):
            logging.info("no overlaps for merging")
            return ""
        paf2gfa = getattr(self.dep, "paf2gfa")

        comm = f"{paf2gfa} -i {fa} -u -c {paf}"
        stdout, stderr = execute(comm)
        # optionally write to file
        if gfa:
            with open(gfa, 'w') as fh:
                fh.write(stdout)
        return stdout


    @staticmethod
    def source_union(edges0: dict[Edge, PafLine], edges1: dict[Edge, PafLine]) -> set:
        """
        Given containment dictionaries (edges), return the union of source nodes.

        :param edges0: First containment dictionary.
        :param edges1: Second containment dictionary
        :return: Union of source nodes.
        """
        sources0 = zip(*edges0.keys())
        sources1 = zip(*edges1.keys())
        try:
            set0 = set(list(sources0)[0])
        except IndexError:
            set0 = set()

        try:
            set1 = set(list(sources1)[0])
        except IndexError:
            set1 = set()

        source_union = set0 | set1
        return source_union




class Sequence:

    def __init__(
        self,
        header: str,
        seq: str,
        cov: Optional[NDArray] = None,
        merged_components: Optional[set[str]] = None,
        merged_atoms: Optional[set[str]] = None,
        cap_l: bool = False,
        cap_r: bool = False,
    ):
        """
        Initialize a Sequence object.

        :param header: The header of the sequence.
        :param seq: The sequence data.
        :param cov: The coverage array (optional).
        :param merged_components: The list of merged components (optional).
        :param merged_atoms: The list of merged atoms (optional).
        :param cap_l: Flag indicating if the sequence has a left cap (default: False).
        :param cap_r: Flag indicating if the sequence has a right cap (default: False).
        """

        self.header = header
        self.seq = seq

        if cov is None:
            self.cov = np.ones(shape=len(seq), dtype='float')
        else:
            self.cov = cov

        # merged_headers provides the info which reads were used to create a new sequence
        if merged_components is None:
            self.atoms = set()  # includes all sequence reads that are contained in this one
            self.components = set()  # all sequences used to build this one
        else:
            # if merged info are provided
            self.components = set(merged_components)
            self.atoms = set(merged_atoms)
        # inits
        self.tetramer_zscores = 0
        # self.kmers = 0
        # temperature for ignoring reads
        # if temperature reaches 0, reads gets frozen
        self.temperature = 30
        self.cap_l = cap_l
        self.cap_r = cap_r
        # accept overlaps
        self.acceptor = True


    def is_hot(self) -> bool:
        """
        Is the sequence active?

        :return: True if temperature is >0, False otherwise.
        """
        if self.temperature > 0:
            return True
        else:
            return False


    def chunk_up_coverage(self, n: int):
        """
        Chunk up the coverage array to reduce resolution of strategies.

        :param n: The chunk size.
        """
        self.cov_chunked = np.array([np.sum(self.cov[i: i + n]) for i in range(0, len(self.cov), n)])
        # init an empty array to record nodes of interest
        self.noi = np.zeros(shape=self.cov_chunked.shape[0], dtype="bool")
        self.scores = np.zeros(shape=self.cov_chunked.shape[0], dtype="float")
        self.benefit = np.zeros(shape=self.cov_chunked.shape[0], dtype="float")


    def contig_scores(self, score_vec: NDArray, n: int = 100):
        """
        Calculate scores for the contig.

        :param score_vec: The precomputed scoring vector.
        :param n: The node size.
        """
        sc = Benefit.score_array(score_vec=score_vec, cov_arr=self.cov_chunked, node_size=n)
        self.scores = sc
        assert self.cov_chunked.shape[0] == self.scores.shape[0]


    def contig_benefits(self, mu: int, ccl: NDArray, node_size: int = 100):
        """
        Calculate benefits for the contig.

        :param mu: Length of anchor bases
        :param ccl: The ccl value.
        :param node_size: The node size.
        """
        benefit, smu_sum = Benefit.calc_fragment_benefit(
            scores=self.scores,
            mu=mu,
            approx_ccl=ccl,
            node_size=node_size,
            e1=bool(self.noi[0]),
            e2=bool(self.noi[-1])
        )
        self.benefit = benefit
        self.smu_sum = smu_sum
        assert self.cov_chunked.shape[0] == self.benefit.shape[1]


    def set_contig_ends(self, n: int, lim: int = 50):
        """
        Declare ends of contig.

        :param n: The node size.
        :param lim: The coverage limit (default: 50).
        """
        cc = self.cov_chunked
        if cc[0] > lim * n:
            pass
        elif self.cap_l:
            pass
        # only set as interesting if every check fails
        # for either model: mark as noi or give max score
        else:
            self.noi[0] = 1
            self.scores[0] = 1

        if cc[-1] > lim * n:
            pass
        elif self.cap_r:
            pass
        else:
            self.noi[-1] = 1
            self.scores[-1] = 1


    def find_strat_m0(self, threshold: float) -> NDArray:
        """
        Find the strategy for the contig using benefit threshold.

        :param threshold: threshold value.
        :return: Strategy array.
        """
        strat = np.where(self.benefit >= threshold, True, False)
        return strat.transpose()




class SequencePool:
    def __init__(
        self,
        sequences: Optional[dict[str, str | Sequence]] = None,
        name: str = "dummy",
        min_len: int = 3000,
        out_dir: str = "dummy",
        threads: int = 6
    ):
        """
        Initialize a SequencePool object. Unified pool for reads and contigs with persistent AVA

        :param sequences: Dictionary of raw sequences. Keys are headers, values are either raw seqs or Sequence objects
        :param name: Name of the sequence pool.
        :param min_len: Minimum length of sequences to be included in the pool. Default is 3000.
        :param out_dir: Output directory.
        :param threads: Number of threads to use. Default is 6.
        """
        self.min_len = min_len
        self.out_dir = out_dir
        self.sequences = dict()
        self.threads = threads
        self.name = name

        if sequences:
            input_type = type(list(sequences.values())[0])
            # raw sequences
            if input_type == str:
                self._ingest_dict(seqs=sequences)
            # Sequence objects
            elif input_type == Sequence:
                self.sequences = sequences
            else:
                print("SequencePool input type not supported")

        self.polished = {}
        # filenames
        self.fa = f'{name}.fa'  # fasta of whole pool
        self.contig_fa = f'{name}.contig.fa'  # fasta of long sequences to map against
        self.ava = f'{name}.ava'  # ava in paf
        self.gfa = f'{name}.gfa'
        self.dep = Dependencies()


    def headers(self) -> set[str]:
        """
        Get the set of sequence headers in the SequencePool.

        :return: Set of sequence headers.
        """
        return set(self.sequences.keys())


    def seqdict(self) -> dict[str, str]:
        """
        Get a dictionary representation of the raw sequences in the SequencePool.

        :return: Dictionary of sequence headers and their corresponding raw sequences.
        """
        return {header: seqo.seq for header, seqo in self.sequences.items()}


    def total_bases(self) -> int:
        """
        Get the total number of bases in the SequencePool.

        :return: Total number of bases.
        """
        return int(np.sum([len(seqo.seq) for header, seqo in self.sequences.items()]))


    def is_empty(self) -> bool:
        """
        Check if the pool is empty

        :return: True if no sequences in pool, False otherwise
        """
        empty = True if len(self.sequences) == 0 else False
        return empty


    def ingest(self, seqs: Any) -> None:
        """
        Add a pile of sequences to the SequencePool.

        :param seqs: Dictionary of raw sequences or an existing SequencePool to add.
        """
        if type(seqs) is dict:
            skipped = self._ingest_dict(seqs=seqs)
            logging.info(f"ingested: {len(seqs) - skipped} pool size: {len(self.sequences.keys())}")
        elif type(seqs) is SequencePool:
            self._ingest_pool(new_pool=seqs)
            logging.info(f"ingested: {len(seqs.sequences)} pool size: {len(self.sequences.keys())}")
        else:
            logging.info("seqs need to be dict, or SequencePool")


    def _ingest_dict(self, seqs: dict[str, str]) -> int:
        """
        Ingest a dictionary of raw sequences.

        :param seqs: Dictionary of raw sequences.
        :return: Number of skipped sequences.
        """
        skipped = 0
        for rid, seq in seqs.items():
            if len(seq) > self.min_len:
                # init sequence object without existing arrays
                seqo = Sequence(header=rid, seq=seq)
                self.sequences[rid] = seqo
            else:
                skipped += 1
        return skipped


    def _ingest_pool(self, new_pool: 'SequencePool') -> None:
        """
        Ingest sequences from an existing SequencePool, e.g. after merging Sequences

        :param new_pool: Existing SequencePool to ingest sequences from.
        """
        for rid, seqo in new_pool.sequences.items():
            if len(seqo.seq) > self.min_len:
                self.sequences[rid] = seqo


    def run_ava(self, sequences: dict[str, str], fa: str, paf: str, base_level: bool = False) -> str:
        """
        Perform All-Versus-All (AVA) comparison of sequences using minimap2.

        :param sequences: Dictionary of sequences.
        :param fa: Filename for temporary sequence file.
        :param paf: Filename for the output PAF file.
        :param base_level: Whether to perform base-level AVA. Default is False.
        :return: Filename of PAF file.
        """
        logging.info(f"Running ava for: {len(sequences)} queries")
        # write current pool to file
        self.write_seq_dict(seq_dict=sequences, file=fa)
        mm2 = getattr(self.dep, "minimap2")

        comm = f'{mm2} -x ava-ont -t{self.threads} {fa} {fa} >{paf}'
        if base_level:
            comm = f'{mm2} -cx ava-ont -t{self.threads} {fa} {fa} >{paf}'

        tic = time.time()
        stdout, stderr = execute(comm)
        toc = time.time()
        logging.info(f"t: {toc - tic}")
        if os.path.exists(f'{self.out_dir}/logs'):
            write_logs(stdout, stderr, f'{self.out_dir}/logs/ava')
        return paf


    def initial_asm_miniasm(self, c: int = 2, trds: int = 8) -> 'SequencePool':
        """
        Perform initial assembly using miniasm.

        :param c: Minimum coverage, fwd to miniasm. Default is 3.
        :param trds: Number of threads. Default is 8.
        :return: SequencePool object containing assembled unitigs.
        """
        sfile = f'{self.name}.init_reads.fa'
        self.write_seq_dict(self.seqdict(), file=sfile)
        mm2 = getattr(self.dep, "minimap2")
        miniasm = getattr(self.dep, "miniasm")

        comm0 = f'{mm2} -x ava-ont -t{trds} {sfile} {sfile} >{sfile}.ava'
        stdout, stderr = execute(comm0)
        write_logs(stdout, stderr, f'{self.out_dir}/contigs/init/init_ava')
        comm1 = f"{miniasm} -f {sfile} {sfile}.ava -c{c} >{self.name}.init.gfa"
        stdout, stderr = execute(comm1)
        write_logs(stdout, stderr, f'{self.out_dir}/contigs/init/init_asm')

        contigs = load_gfa(f'{self.name}.init.gfa')
        contig_pool = SequencePool(contigs)
        # disallow extension of circular contigs
        for header, seqo in contig_pool.sequences.items():
            if header[-1] == 'c':
                seqo.acceptor = False
        return contig_pool


    def add2ava(self, new_sequences: 'SequencePool') -> tuple[str, str]:
        """
        Add new sequences to the All-Versus-All (AVA) comparisons.
        Instead of rerunning complete ava, run ava for new sequences and map new sequences onto existing ones

        :param new_sequences: SequencePool object containing the new sequences.
        :return: Filenames of the new AVA file and the file containing mappings of new seqs to previous ones.
        """
        logging.info(f'adding to ava: {len(new_sequences.sequences)}')
        self.write_seq_dict(seq_dict=self.seqdict(), file=self.fa)
        # declare new filenames
        new_fa = f'{self.fa}.new'
        new_ava = f'{self.ava}.new'
        new_onto_pool = f'{self.fa}.new.onto_pool'
        # write new reads to file
        self.write_seq_dict(seq_dict=new_sequences.seqdict(), file=new_fa)
        # ava of new sequences
        mm2 = getattr(self.dep, "minimap2")
        comm = f'{mm2} -x ava-ont -t{self.threads} {new_fa} {new_fa} >{new_ava}'
        stdout, stderr = execute(comm)
        write_logs(stdout, stderr, f'{self.out_dir}/logs/ava_add')
        # mapping new sequences to previous pool
        comm = f'{mm2} -x map-ont -w5 -e0 -m100 -r2k -t{self.threads} {self.fa} {new_fa} >{new_onto_pool}'
        stdout, stderr = execute(comm)
        write_logs(stdout, stderr, f'{self.out_dir}/logs/map2pool')
        # return filenames to be ingested as AVA
        return new_ava, new_onto_pool


    def remove_sequences(self, sequences: set[str]) -> None:
        """
        Remove sequences from the pool based on their IDs.
        Once a read is used, it can not contribute to other alignments

        :param sequences: Set of sequence IDs to remove.
        """
        pre = len(self.sequences)
        popped = 0
        for sid in sequences:
            self.sequences.pop(sid, None)
            popped += 1
        post = len(self.sequences)
        logging.info(f'Removed: {popped} ({pre} {post})')


    def trim_sequences(self, trim_dict: dict[str, tuple[int, int, str]]) -> dict[str, str]:
        """
        Trim sequences in order to create valid alignments.

        :param trim_dict: Dictionary of [sid: (start:stop, other_name)] for trimming sequences.
        :return: Dictionary of trimmed sequences.
        """
        trimmed_seqs = {}
        other_seqs = {}
        valid_ids = set()

        for sid, (start, stop, other) in trim_dict.items():
            try:
                nsid = sid + '%'
                trimmed_seqs[nsid] = deepcopy(self.sequences[sid])
                other_seqs[other] = self.sequences[other]
                valid_ids.add(nsid)
            except KeyError:
                logging.info("key for trimming not in sequence pool")
                continue

        for sid, (start, stop, other) in trim_dict.items():
            nsid = sid + '%'
            # skip sequences that were not found in previous step
            if nsid not in valid_ids:
                continue
            seqo = trimmed_seqs[nsid]
            # deselect the trimmed bit with a boolean mask
            mask = np.ones(shape=len(seqo.seq), dtype='bool')
            mask[start: stop] = 0
            seq_arr = np.array(list(seqo.seq))
            seq_arr_trim = seq_arr[mask]
            seq_trim = ''.join(seq_arr_trim)
            # replace attributes of the Sequence Obj
            seqo.seq = seq_trim
            seqo.cov = seqo.cov[mask]
            seqo.header = nsid

        trimmed_pool = SequencePool(sequences=trimmed_seqs)
        other_pool = SequencePool(sequences=other_seqs)
        # ingest pool of trimmed sequences
        self.ingest(seqs=trimmed_pool)

        # combine sequence dicts for mapping
        seq_dict = dict(trimmed_pool.seqdict(), **other_pool.seqdict())
        return seq_dict


    @staticmethod
    def get_next_increment_edges(edges: set[Edge], previous_edges: set[Edge] = None) -> tuple[set[Edge], set[Edge]]:
        """
        Get the next increment edges based on the given edges and previous edges.
        if no previous edges are given, get the edges with in-degree of 0: This is the start of the algorithm

        :param edges: Set of edges.
        :param previous_edges: Set of previous edges.
        :return: Tuple containing the remaining edges and the next increment edges.
        """
        if not previous_edges:
            sources, targets = zip(*edges)
            next_sources = set(sources) - set(targets)
        else:
            # otherwise grab the edges starting at the previous targets
            next_sources = [t for (s, t) in previous_edges]
        # get the next edges to increment
        next_edges = {(s, t) for (s, t) in edges if s in next_sources}
        # remove the next edges
        for e in next_edges:
            edges.remove(e)
        return edges, next_edges


    def effect_increment(self, source: str, target: str, rec: PafLine, edge_multiplicity: float) -> None:
        """
        Effect the increment by adjusting the coverage and adding the source as a constituent of the target.

        :param source: Source sequence ID.
        :param target: Target sequence ID.
        :param rec: Mapping record.
        :param edge_multiplicity: Edge multiplicity.
        """
        # grab relevant coordinates of this containment
        ostart, oend, olen, cstart, cend, clen = rec.grab_increment_coords()
        # grab source coverage
        cont_cov = self.sequences[source].cov[cstart: cend].copy()
        # adjust length of coverage array
        if clen == olen:
            pass
        elif clen > olen:
            cont_cov = cont_cov[: olen]
        elif clen < olen:
            cont_cov = np.pad(cont_cov, (0, olen - clen), mode='edge')

        # account for reverse complement
        if rec.rev:
            cont_cov = cont_cov[::-1]
        else:
            pass

        # adjust for edge multiplicity
        cont_cov /= edge_multiplicity

        # add the source coverage to target coverage
        self.sequences[target].cov[ostart: oend] += cont_cov
        # limit coverage to 100 to prevent mess
        self.sequences[target].cov[np.where(self.sequences[target].cov > 100)] = 100

        # add the source as constituent of target
        if '*' not in source:
            self.sequences[target].atoms.add(source)


    def effect_increments(self, next_edges: set[Edge], containment: dict[Edge, PafLine], edge_multiplicity: dict[str, float] = None) -> None:
        """
        Effect the increments based on the next set of edges.

        :param next_edges: Set of next edges.
        :param containment: Containment records.
        :param edge_multiplicity: Edge multiplicity.
        """
        # loop over the next set of edges to effect the collected increments
        for (source, target) in next_edges:
            rec = containment[(source, target)]
            # grab edge multiplicity modifier, in case of parallel edges
            if edge_multiplicity is None:
                em = 1
            else:
                em = edge_multiplicity[source]
            self.effect_increment(source, target, rec, em)


    @staticmethod
    def find_edge_multiplicity(edges: set[Edge]):
        """
        Find the edge multiplicity for the given edges.

        :param edges: Set of edges.
        """
        sources, targets = zip(*edges)
        source_counts = Counter(sources)
        return source_counts


    def increment(self, containment: dict[Edge, PafLine]) -> set:
        """
        Increment the coverage counts based on containment records.

        :param containment: Containment records.
        :return: Set of IDs of the contained sequences.
        """
        edges = set(containment.keys())
        if not edges:
            # nothing to do
            return set()

        # debugging
        # import sys
        # sys.path.insert(0, "/home/lukas/Desktop/Aeons/code/plot")
        # from gt_plot import containment_graph
        # containment_graph(edges, 0)

        # get the first edges to increment, i.e. those with 0 in-degree
        edges, next_edges = self.get_next_increment_edges(edges, previous_edges=None)
        if not next_edges:
            return set()
        edge_multiplicity = self.find_edge_multiplicity(next_edges)
        self.effect_increments(next_edges, containment, edge_multiplicity)
        previous_edges = next_edges

        while len(edges) > 0:
            # get the next edges
            edges, next_edges = self.get_next_increment_edges(edges, previous_edges=previous_edges)
            if not next_edges:
                return set()
            edge_multiplicity = self.find_edge_multiplicity(next_edges)
            self.effect_increments(next_edges, containment, edge_multiplicity)

            # circular containment relationships could trap us here
            if len(next_edges) == len(previous_edges):
                break
            previous_edges = next_edges

        # for removal: return the ids of the contained sequences
        contained_ids = set([s for (s, t) in containment.keys()])
        return contained_ids


    def reset_temperature(self, sids: set[str], t: int = 50) -> None:
        """
        Reset the temperature of active reads to the given value.

        :param sids: Set of sequence IDs.
        :param t: Temperature value.
        """
        # give active reads a boost in temperature
        for s in sids:
            try:
                seqo = self.sequences[s]
                seqo.temperature = t
            except KeyError:
                pass


    def decrease_temperature(self, lim: int) -> set[str]:
        """
        Decrease the temperature of all reads and return the frozen sequences.
        Sequences longer than lim never freeze

        :param lim: Length limit.
        :return: Set of frozen sequence headers.
        """
        frozen_seqs = set()
        for header, seqo in self.sequences.items():
            if len(seqo.seq) < lim:
                seqo.temperature -= 1
                if not seqo.is_hot():
                    frozen_seqs.add(header)
        logging.info(f"frozen seqs: {len(frozen_seqs)}")
        return frozen_seqs


    def declare_contigs(self, min_contig_len: int) -> 'SequencePool':
        """
        Declare contigs based on a minimum contig length. These will be used for strategy generation

        :param min_contig_len: Minimum contig length.
        :return contig_pool: Pool of contigs
        """
        contigs = {header: seqo for header, seqo in self.sequences.items() if len(seqo.seq) > min_contig_len}
        contig_pool = SequencePool(sequences=contigs)
        return contig_pool


    def has_min_one_contig(self, min_contig_len: int) -> bool:
        """
        Declare contigs based on a minimum contig length. These will be used for strategy generation

        :param min_contig_len: Minimum contig length.
        :return contig: bool
        """
        contigs = {header: seqo for header, seqo in self.sequences.items() if len(seqo.seq) > min_contig_len}
        contig = True if contigs else False
        return contig


    def get_atoms(self, headers: list) -> set[str]:
        """
        Get all atomic reads based on the given headers.

        :param headers: List of headers.
        :return: Set of atomic reads.
        """
        # given a list of headers, get all atomic reads
        atoms = set()
        for h in headers:
            atm = self.sequences[h].atoms
            atoms.update(atm)
        return atoms


    def get_components(self, headers: list) -> set[str]:
        """
        Get all component reads based on the given list of headers.

        :param headers: List of headers.
        :return: Set of component reads.
        """
        components = set()
        for h in headers:
            cmp = self.sequences[h].components
            components.update(cmp)
            components.add(h)
        return components


    @staticmethod
    def write_seq_dict(seq_dict: dict[str, str], file: str) -> None:
        """
        Write a sequence dictionary to file.

        :param seq_dict: Sequence dictionary.
        :param file: File to write the dictionary to.
        """
        with open(file, 'w') as fasta:
            for sid, seq in seq_dict.items():
                fasta.write(f'>{sid}\n')
                fasta.write(f'{seq}\n')



    @staticmethod
    def load_unitigs(gfa: str) -> list['Unitig']:
        """
        Load unitigs after graph cleaning with gfatools.
        Input can be a gfa file or the content of the file

        :param gfa: Path to the GFA file or a string representation of the GFA.
        :return: A list of Unitig objects.
        """
        if not gfa:
            return []

        if gfa.startswith('S\t'):
            gfat = gfa
        else:
            with open(gfa, 'r') as fh:
                gfat = fh.read()
        # string operations
        gfatt = gfat.split('\nS\t')
        # first and last need extra treatment
        gfatt[0] = gfatt[0][2:]  # skip the first "S\t"
        gfatt[-1] = gfatt[-1].replace('\nx', '\nL').split('\nL')[0]  # exclude all L and x lines
        # get the x lines
        gfax = gfat.split('\nx\t')[1:]
        # make sure we have the same number of S_A lines and x lines
        assert len(gfatt) == len(gfax)
        # parse into unitig objects
        unitigs = []
        for sa_lines, x_line in zip(gfatt, gfax):
            unitigs.append(Unitig(sa_lines.split('\n'), x_line))
        return unitigs


    def is_intra(self, seq1: str, seq2: str) -> bool:
        """
        Check if two sequences are classified as intraspecific based on the Euclidean distance of tetramers.

        :param seq1: Name of first sequence.
        :param seq2: Name of second sequence.
        :return: True if the sequences are classified as intraspecific, False otherwise.
        """
        s1 = self.sequences[seq1]
        s2 = self.sequences[seq2]
        euc = euclidean_dist(s1, s2)
        return euc < euclidean_threshold



class ContigPool(SequencePool):


    def process_contigs(
        self,
        score_vec: NDArray,
        ccl: NDArray,
        out_dir: str,
        lam: float,
        batch: int,
        mu: int = 400,
        node_size: int = 100,
    ) -> dict[str, NDArray]:
        """
        WRAPPER to process contigs using m0 strategy.

        :param score_vec: The score vector.
        :param node_size: The node size.
        :param ccl: Read length distribution.
        :param out_dir: The output directory.
        :param mu: Length of anchor bases.
        :param lam: The lam value.
        :param batch: The batch number.
        :return: A dictionary containing the new strategies for contigs.
        """
        logging.info("finding new strategies..")
        self._chunk_up_contigs(node_size=node_size)
        self._contigs_scores(score_vec=score_vec, node_size=node_size)
        self._process_contig_ends(node_size=node_size)
        self._contigs_benefits(ccl=ccl, mu=mu, node_size=node_size)
        t = self.find_threshold(mu=mu, lam=lam, node_size=node_size)
        # find and write new strategies
        contig_strats = self._find_contig_strategies(t=t)
        logging.info("writing new strategies")
        self._write_contig_strategies(out_dir=out_dir, contig_strats=contig_strats)
        self._write_index_file(out_dir=out_dir, batch=batch)
        return contig_strats


    def _chunk_up_contigs(self, node_size: int = 100) -> None:
        """
        Chunk up collection of contigs by giving them a chunked representation to reduce resolution

        :param node_size: The node size.
        """
        n_comp = 0
        lengths = []
        for header, seqo in self.sequences.items():
            seqo.chunk_up_coverage(n=node_size)
            n_comp += 1
            lengths.append(len(seqo.seq))
        lengths_sort = np.sort(lengths)[::-1]
        logging.info(f'num components: {n_comp}')
        logging.info(f'total comp length: {lengths_sort.sum()}')
        logging.info(f'longest components: {lengths_sort[:10]}')


    def _contigs_scores(self, score_vec: NDArray, node_size: int = 100) -> None:
        """
        Get the scores for each contig.

        :param score_vec: The score vector.
        :param node_size: The node size.
        """
        for header, seqo in self.sequences.items():
            seqo.contig_scores(score_vec=score_vec, n=node_size)


    def _contigs_benefits(self, ccl: NDArray, mu: int, node_size: int = 100) -> None:
        """
        Get the benefit for each contig.

        :param ccl: Read length distribution
        :param mu: The length of anchor bases.
        :param node_size: The node size.
        """
        for header, seqo in self.sequences.items():
            seqo.contig_benefits(mu=mu, ccl=ccl, node_size=node_size)


    def find_threshold(self, mu: float, lam: float, node_size: int = 100) -> float:
        """
        Find the acceptance threshold based on all benefit values.

        :param mu: Length of anchor bases
        :param lam: The lam value.
        :param node_size: The node size.
        :return: Acceptance threshold.
        """
        # flatten all benefit values
        benefit = np.column_stack([seqo.benefit for seqo in self.sequences.values()]).ravel()
        smu_sum = np.sum([seqo.smu_sum for seqo in self.sequences.values()])
        # find acceptance threshold
        alpha = 200 // node_size
        rho = 300 // node_size
        tc = (lam - mu - 300) // node_size

        benefit_bin, counts = Benefit.benefit_bins(benefit)
        # average benefit of strategy in the case that all fragments are rejected
        ubar0 = smu_sum
        tbar0 = alpha + rho + (mu // node_size)
        # cumsum of the benefit (bins multiplied by how many sites are in the bin)
        cs_u = np.cumsum(benefit_bin * counts) + ubar0
        cs_t = np.cumsum(tc * counts) + tbar0
        peak = cs_u / cs_t
        strat_size = np.argmax(peak) + 1
        # plt.plot(cs_u)
        # plt.plot(cs_t)
        # plt.plot(peak)
        # plt.show()
        # calculate threshold exponent and where values are geq
        try:
            threshold = float(benefit_bin[strat_size])
        except IndexError:
            threshold = float(benefit_bin[-1])
        return threshold


    def _process_contig_ends(self, node_size: int = 100) -> None:
        """
        Set contig ends to reflect their status during calculation of the strategy.

        :param node_size: The node size.
        """
        for header, seqo in self.sequences.items():
            seqo.set_contig_ends(n=node_size)


    def _find_contig_strategies(self, t: float = 0.0) -> dict[str, NDArray]:
        """
        Find the strategies for all contigs.

        :param t: The score threshold. Defaults to 0.0.
        :return: A dictionary containing the new strategies for contigs.
        """
        contig_strats = {}
        for header, seqo in self.sequences.items():
            cstrat = seqo.find_strat_m0(threshold=t)
            contig_strats[header] = cstrat
        return contig_strats


    @staticmethod
    def _write_contig_strategies(out_dir: str, contig_strats: dict[str, NDArray]) -> None:
        """
        Write the strategies for all contigs to a single file.

        :param out_dir: The output directory.
        :param contig_strats: A dictionary containing the strategies for contigs.
        """
        cpath_tmp = f'{out_dir}/masks/boss_tmp.npz'
        np.savez(cpath_tmp, **contig_strats)
        # after writing to tmpfile, rename to replace the current strat
        cpath = f'{out_dir}/masks/boss.npz'
        Path(cpath_tmp).rename(cpath)
        # Example how to load these:
        # container = np.load(f'{cpath}.npz')
        # data = {key: container[key] for key in container}


    def _write_index_file(self, out_dir: str, batch: int) -> None:
        """
        Write a new contig file to map against. Readfish generates the mmi

        :param out_dir: The output directory.
        :param batch: The batch number.
        """
        fa_path_tmp = f'{out_dir}/contigs/aeons_tmp.fa'
        # save the contigs to fasta
        with open(fa_path_tmp, 'w') as fasta:
            for sid, seqo in self.sequences.items():
                fasta.write(f'>{sid}\n')
                fasta.write(f'{seqo.seq}\n')
        # rename file after being written to
        fa_path = f'{out_dir}/contigs/aeons.fa'
        Path(fa_path_tmp).rename(fa_path)
        # copy previous contigs
        if batch % 10 == 0:
            copy(fa_path, f'{out_dir}/contigs/prev/aeons_{batch}.fa')


class Unitig:

    def __init__(self, sa_line_list: list[str], x_line: str):
        """
        Initialize Unitig object.

        :param sa_line_list: A list of lines from a GFA file corresponding to the S & A lines of a single unitig.
        :param x_line: The X line from the GFA file corresponding to the unitig.
        """
        self.name = random_id()
        self._format_sa(sa_line_list)
        self._format_x(x_line)
        self.cov = None


    def _format_x(self, x_line: str) -> None:
        """
        Format the X line of the unitig.

        :param x_line: The X line from the GFA file.
        """
        xl = x_line.split("\t")
        assert xl[0].startswith('utg')
        try:
            cl = int(xl[3])
            cr = int(xl[4])
        except IndexError:
            self.cap_l = False
            self.cap_r = False
            return
        self.cap_l = True if cl > 0 else False
        self.cap_r = True if cr > 0 else False



    def _format_sa(self, line_list: list[str]) -> None:
        """
        Format the S & A lines of the unitig.

        :param line_list: A list of lines from the GFA file corresponding to the S & A lines of the unitig.
        """
        self._format_sline(sline=line_list[0])
        self._format_atoms(alines=line_list[1:])


    def _format_sline(self, sline: str) -> None:
        """
        Format the S line of the unitig.

        :param sline: The S line from the GFA file.
        """
        sls = sline.split('\t')
        self.seq = sls[1]
        self.length = int(sls[2].split(':')[-1])
        circ = sls[0][-1]
        self.circ = True if circ == 'c' else False
        assert sls[0].startswith('utg')
        assert self.length == len(self.seq)


    def _format_atoms(self, alines: list[str]) -> None:
        """
        Format the A lines (atoms) of the unitig.

        :param alines: A list of lines
        """
        atoms = []
        for line in alines:
            assert line.startswith('A')
            atom = {}
            al = line.split('\t')
            atom['pos'] = int(al[2])
            atom['strand'] = al[3]
            if atom['strand'] == '-':
                atom['rev'] = 1
            elif atom['strand'] == '+':
                atom['rev'] = 0
            else:
                print(line)
                print(al)
                print(atom)
                logging.info("wrong strand spec of unitig")
                exit(1)
            atom['name'] = al[4]
            atoms.append(atom)
        # loop a second time to get the ends
        cpos = 0
        for i in range(len(alines) - 1):
            line = alines[i + 1]
            al = line.split('\t')
            pos = int(al[2])
            to_add = pos - cpos
            atoms[i]['n'] = to_add
            cpos = pos
        # mark last atom
        atoms[-1]['n'] = -1
        self.atoms = atoms
        self.atom_headers = [a['name'] for a in atoms]


    def to_seqo(self, seqpool: SequencePool) -> Sequence:
        """
        Transform unitig into a sequence object. Needs to be done after merging coverage.

        :param seqpool: The sequence pool of the source sequence that made the unitig.
        :return: A sequence object representing the unitig.
        """
        # check that coverage merging has been done
        assert self.cov is not None
        # grab atoms of atoms
        merged_atoms = seqpool.get_atoms(headers=self.atom_headers)
        merged_components = seqpool.get_components(headers=self.atom_headers)
        seqo = Sequence(header=self.name, seq=self.seq, cov=self.cov,
                        merged_components=merged_components, merged_atoms=merged_atoms,
                        cap_l=self.cap_l, cap_r=self.cap_r)
        return seqo



class UnitigPool:

    def __init__(self, unitigs: list[Unitig]):
        """
        Initialize a UnitigPool object.

        :param unitigs: A list of Unitig objects.
        """
        self.unitigs = unitigs


    def get_unitig_coverage_arrays(self, seqpool: SequencePool) -> None:
        """
        Perform coverage array merging for each unitig in the pool.

        :param seqpool: The sequence pool object.
        """
        for u in self.unitigs:
            cm = CoverageMerger(u, seqpool.sequences)
            cov_arr = cm.cov_arr
            u.cov = cov_arr


    def unitigs2seqpool(self, seqpool: SequencePool, min_seq_len: int) -> tuple[SequencePool, set[str]]:
        """
        Transform all unitigs in the pool to sequence objects and obtain the read IDs to remove.

        :param seqpool: The sequence pool object.
        :param min_seq_len: The minimum sequence length for the new pool.
        :return: A tuple containing the new sequence pool and the set of used read IDs.
        """
        seqos = {}
        used_sids = set()
        for u in self.unitigs:
            unitig_seqo = u.to_seqo(seqpool)
            seqos[u.name] = unitig_seqo
            used_sids.update(u.atom_headers)
        # construct a new pool
        new_pool = SequencePool(sequences=seqos, min_len=min_seq_len)
        return new_pool, used_sids



class CoverageMerger:


    def __init__(self, unitig: Unitig, seqpool: dict[str, Sequence]) -> None:
        """
        Initializes a CoverageMerger object.

        :param unitig: Unitig object.
        :param seqpool: Dictionary of sequences used to create the unitigs
        :raises AssertionError: If the length of the unitig is not equal to the shape of the merged coverage array.

        """
        self.unitig = unitig
        cov_arr = self._create_merged_arr(atoms=unitig.atoms, seqpool=seqpool)
        assert unitig.length == cov_arr.shape[0]
        self.cov_arr = cov_arr


    def _create_merged_arr(self, atoms: list[dict[str, Any]], seqpool: dict[str, Sequence]) -> NDArray:
        """
        Creates merged coverage array for the unitig.

        :param atoms: List of atom dictionaries.
        :param seqpool: Dictionary of sequences used to create the unitigs
        :return: The merged coverage array.
        """
        arr_parts = []
        cpos = 0
        for a in atoms:
            assert a['pos'] >= cpos
            name = a['name']
            atom_arr = seqpool[name].cov.copy()
            atom_arr = atom_arr[::-1] if a['rev'] else atom_arr
            if not a['n'] == -1:
                atom_arr = atom_arr[: a['n']]
            else:
                # add last atom (goes until end of atom)
                if not self.unitig.circ:
                    atom_arr = atom_arr
                else:
                    diff = self.unitig.length - a['pos']
                    atom_arr = atom_arr[:diff]
            arr_parts.append(atom_arr)
            cpos = a['pos']
        return np.concatenate(arr_parts)




class MultilineContainments:


    def __init__(self, records: list[PafLine]) -> None:
        """
        Initializes a MultilineContainments object.

        :param records: A list of PafLine objects.
        """
        self.records = records
        # collect all partners with multiple mappings
        multidict = self._fill_multidict()
        # get a dict of (contained, container): [recs]
        containments = self._get_multiline_containments(multidict=multidict)
        self.containments = containments


    def _fill_multidict(self) -> dict[str, list[PafLine]]:
        """
        Fills a multidict with records that have multiple mappings.

        :return: The multidict of (key, records) pairs.
        """
        multidict = defaultdict(list)
        for rec in self.records:
            multidict[rec.keygen()].append(rec)
        multidict = {k: recs for k, recs in multidict.items() if len(recs) > 1}
        return multidict


    def _get_multiline_containments(self, multidict: dict[str, list[PafLine]]) -> dict[Edge, PafLine]:
        """
        Gets the multiline containments from the multidict using multiple internal mappings.

        :param multidict: The multidict of (key, records) pairs.
        :return: Dictionary of (contained, container) tuples mapped to PafLine objects.
        """
        containments = {}
        for k, recs in multidict.items():
            cont = self.multiline_containment(recs)
            if cont:
                containments.update(cont)
        return containments


    @staticmethod
    def multiline_containment(records: list[PafLine], n: int = 100) -> dict[Edge, PafLine]:
        """
        Checks for multiline containment from multiple internal match mappings.

        :param records: A list of PafLine objects.
        :param n: Node size, defaults to 100.
        :return: The dictionary of (contained, container) tuples mapped to PafLine objects, or empty dict
        """
        qlen = records[0].qlen // n
        tlen = records[0].tlen // n
        qarr = np.zeros(shape=qlen , dtype="bool")
        tarr = np.zeros(shape=tlen , dtype="bool")

        if len(records) > 10:
            return {}

        for r in records:
            qarr[r.qstart // n: r.qend // n] = 1
            tarr[r.tstart // n: r.tend // n] = 1

        # if more than 0.9 are covered by mappings
        if sum(qarr) > qlen * 0.9:
            # if overhang is smaller than 0.15 of len
            q_low, q_high = np.nonzero(qarr)[0][[0, -1]]
            if (q_high - q_low) > qlen * 0.85:
                # return containment tuple
                t_low, t_high = np.nonzero(tarr)[0][[0, -1]]
                ctd_low, ctd_high, ctr_low, ctr_high = q_low, q_high, t_low, t_high
                cont = MultilineContainments.generate_paf_cont(
                    records, 'q', 't', ctd_low, ctd_high, ctr_low, ctr_high, n
                )
                return cont

        if sum(tarr) > tlen * 0.9:
            t_low, t_high = np.nonzero(tarr)[0][[0, -1]]
            if (t_high - t_low) > tlen * 0.85:
                q_low, q_high = np.nonzero(qarr)[0][[0, -1]]
                ctd_low, ctd_high, ctr_low, ctr_high = t_low, t_high, q_low, q_high
                cont = MultilineContainments.generate_paf_cont(
                    records, 't', 'q', ctd_low, ctd_high, ctr_low, ctr_high, n
                )
                return cont

        # if neither q nor t are contained
        return {}


    @staticmethod
    def generate_paf_cont(
        records: list[PafLine],
        ctd: str,
        ctr: str,
        ctd_low: int,
        ctd_high: int,
        ctr_low: int,
        ctr_high: int,
        n: int
    ) -> dict[Edge, PafLine]:
        """
        Generates a PafLine object that describes the containment,
         given a list of records that make up one multicontainment

        :param records: A list of PafLine objects.
        :param ctd: The contained sequence name.
        :param ctr: The container sequence name.
        :param ctd_low: The lower bound of the contained sequence.
        :param ctd_high: The upper bound of the contained sequence.
        :param ctr_low: The lower bound of the container sequence.
        :param ctr_high: The upper bound of the container sequence.
        :param n: Node size.
        :return: The dictionary of (contained, container) tuples mapped to PafLine objects.
        """
        ctd_name = getattr(records[0], f'{ctd}name')
        ctr_name = getattr(records[0], f'{ctr}name')
        ctd_len = getattr(records[0], f'{ctd}len')
        ctr_len = getattr(records[0], f'{ctr}len')
        # use maximum span
        ctd_span = ctd_high - ctd_low
        ctr_span = ctr_high - ctr_low
        # except if span on container longer than 2.2x of contained
        if ctr_span > 2.2 * ctd_span:
            r = 0
            maplen = 0
            for i in range(len(records)):
                # then use the longest alignment between the two
                if records[i].maplen > maplen:
                    maplen = records[i].maplen
                    r = i
            ctr_low = getattr(records[r], f'{ctr}start') // n
            ctr_high = getattr(records[r], f'{ctr}end') // n
        # generate paf entry
        paf = f'{ctd_name}\t{ctd_len}\t{ctd_low * n}\t{ctd_high * n}\t+' \
              f'\t{ctr_name}\t{ctr_len}\t{ctr_low * n}\t{ctr_high * n}\t0\t0\t0'
        rec = PafLine(paf)
        # mark as first contained
        rec.c = 2
        return {(ctd_name, ctr_name): rec}




class Benefit:

    @staticmethod
    def init_scoring_vec(lowcov: float) -> NDArray:
        """
        Initialize scoring vector based on target coverage.

        :param lowcov: The target coverage value.
        :return: The scoring vector.
        """
        x = np.arange(101)
        # a = lowcov * 5
        # score_vec = -gamma.cdf(x, a=a, scale=0.2) + 1
        score_vec = 1 / (np.exp(x - lowcov) + 1)
        return score_vec


    @staticmethod
    def score_array(score_vec: NDArray, cov_arr: NDArray, node_size: int = 100) -> NDArray:
        """
        Calculate scores based on the scoring vector and coverage array.

        :param score_vec: scoring vector.
        :param cov_arr: coverage array.
        :param node_size: node size.
        :return: The calculated scores.
        """
        # grab scores using multi-indexing
        carr = cov_arr // node_size  # apply resolution reduction
        carr_int = carr.astype("int")
        scores = score_vec[carr_int]
        return scores


    @staticmethod
    def calc_fragment_benefit(
        scores: NDArray,
        mu: int,
        approx_ccl: NDArray,
        e1: bool,
        e2: bool,
        node_size: int = 100,
    ) -> tuple[NDArray, float]:
        """
        Calculate the benefit of a fragment based on scores, mu, node_size, approx_ccl, e1, and e2.

        :param scores: Fragment's position-wise scores.
        :param mu: Length of anchor bases.
        :param node_size: node size.
        :param approx_ccl: Approx of read length distribution.
        :param e1: Left-end marker.
        :param e2: Right-end marker.
        :return: The calculated benefit and smu_sum.
        """
        # expand score to account for contig ends
        mu_ds = mu // node_size
        ccl_ds = approx_ccl // node_size
        ccl_max = int(ccl_ds[-1])
        sx = Benefit._expand_scores(scores, e1, e2, ccl_max)
        smu = Benefit._calc_smu_moving(score=sx, mu_ds=mu_ds)
        benefit = Benefit._calc_benefit_moving(score=sx, ccl_ds=ccl_ds)
        smu_sum = float(np.sum(smu))
        b = benefit - smu
        b[b < 0] = 0
        b = b[:, ccl_max: -ccl_max]
        assert b.shape[1] == scores.shape[0]
        return b, smu_sum


    @staticmethod
    def _expand_scores(scores: NDArray, e1: bool, e2: bool, ccl_max: int) -> NDArray:
        """
        Expand scores to account for contig ends.

        :param scores: Fragment scores.
        :param e1: Left-end marker.
        :param e2: Right-end marker.
        :param ccl_max: Max of approx read length dist.
        :return: The expanded scores.
        """
        scoresx = np.zeros(shape=scores.shape[0] + (ccl_max * 2), dtype="float64")
        scoresx[ccl_max: -ccl_max] = scores
        scoresx[0: ccl_max] = 1 if e1 else 0
        scoresx[-ccl_max: -1] = 1 if e2 else 0
        return scoresx


    @staticmethod
    def _calc_smu_moving(score: NDArray, mu_ds: int) -> NDArray:
        """
        Calculate smu moving based on score and down-sampled mu.

        :param score: The score.
        :param mu_ds: The mu_ds.
        :return: The calculated smu.
        """
        smu_fwd = bn.move_sum(score, window=mu_ds, min_count=1)
        smu_rev = bn.move_sum(score[::-1], window=mu_ds, min_count=1)
        smu = np.stack((smu_fwd, smu_rev))
        return smu


    @staticmethod
    def _calc_benefit_moving(score: NDArray, ccl_ds: NDArray) -> NDArray:
        """
        Calculate benefit moving based on score and ccl_ds.

        :param score: Fragment scores.
        :param ccl_ds: Down-sampled read length dist array.
        :return: The calculated benefit.
        """
        score_rev = score[::-1]
        benefit = np.zeros(shape=(2, score.shape[0]), dtype="float64")
        perc = np.arange(0.1, 1.1, 0.1)[::-1]
        assert perc.shape == ccl_ds.shape
        for i in range(ccl_ds.shape[0]):
            ben_fwd = bn.move_sum(score, window=int(ccl_ds[i]), min_count=1)[ccl_ds[i]: -1]
            ben_rev = bn.move_sum(score_rev, window=int(ccl_ds[i]), min_count=1)[ccl_ds[i]: -1]
            benefit[0, 0: -ccl_ds[i] - 1] += ben_fwd * perc[i]
            benefit[1, ccl_ds[i]: -1] += ben_rev[::-1] * perc[i]
        return benefit


    @staticmethod
    def benefit_bins(benefit: NDArray) -> tuple[NDArray, NDArray]:
        """
        Group benefit into bins of similar values using binary exponent. Used to find acceptance threshold

        :param benefit: positional benefit array.
        :return: The benefit bins and counts.
        """
        benefit_nz_ind = np.nonzero(benefit)
        benefit_flat_nz = benefit[benefit_nz_ind]
        # to make binary exponents work, normalise benefit values
        normaliser = np.max(benefit_flat_nz)
        benefit_flat_norm = benefit_flat_nz / normaliser
        mantissa, benefit_exponents = np.frexp(benefit_flat_norm)
        # count how often each exponent is present
        # absolute value because counting positive integers is quicker
        benefit_exponents_pos = np.abs(benefit_exponents)
        # multi-thread counting of exponents
        exponent_arrays = np.array_split(benefit_exponents_pos, 12)
        with TPexe(max_workers=12) as executor:
            exponent_counts = executor.map(np.bincount, exponent_arrays)
        exponent_counts = list(exponent_counts)
        # aggregate results from threads
        # target array needs to have the largest shape of the thread results
        max_exp = np.max([e.shape[0] for e in exponent_counts])
        bincounts = np.zeros(shape=max_exp, dtype='int')
        # sum up results from individual threads
        for exp in exponent_counts:
            bincounts[0:exp.shape[0]] += exp
        # filter empty bins
        exponents_unique = np.nonzero(bincounts)[0]
        # counts of the existing benefit exponents
        counts = bincounts[exponents_unique]
        # use exponents to rebuild benefit values
        benefit_bin = np.power(2.0, -exponents_unique) * normaliser
        return benefit_bin, counts







from itertools import product
from math import sqrt
from typing import Any

import numpy as np
from numpy.typing import NDArray

from boss.utils import reverse_complement



class KmerCounter:

    def __init__(self):
        """
        Initialize the KmerCounter object.
        Specifically for di-, tri-, and tetramers
        General purpose class that can be initialized once, then used for many sequences
        """
        # initialise kmer_tables for n = {2, 3, 4}
        self.t2, self.k2 = self._kmer_table(2)
        self.t3, self.k3 = self._kmer_table(3)
        self.t4, self.k4 = self._kmer_table(4)


    @staticmethod
    def _kmer_table(k: int) -> tuple[NDArray, list[str]]:
        """
        Generate a kmer table.

        :param k: The length of the kmers.
        :return: The kmer table and the list of kmers as strings.
        """
        # initialise nucleotides and their ordinants
        NUC = ['A', 'C', 'G', 'T']
        ordinants = [ord(c) for c in NUC]
        # generate all possible kmers for k
        kmers_str = [''.join(nu) for nu in product(NUC, repeat=k)]
        # the respective integer kmers
        # numbers modified to reduce table size
        kmers = np.array(list(product(ordinants, repeat=k))) - 64
        kmers = np.clip(kmers, 1, 8)
        # create multi-dim array for indexing
        ishape = [np.max(kmers) + 1] * k
        table = np.zeros(shape=ishape, dtype=np.uint8)
        # populate table
        kmer_ind = np.arange(len(kmers))
        for j in range(len(kmers)):
            # using tuple to NOT trigger fancy indexing
            table[tuple(kmers[j])] = kmer_ind[j]
        return table, kmers_str


    @staticmethod
    def _integer_seq(seq: str) -> NDArray:
        """
        Convert a sequence into an integer array.

        :param seq: The input sequence.
        :return: The integer array representing the sequence.
        """
        # prepare a sequence with its reverse complement in byte array
        pseq = np.array([seq + reverse_complement(seq)], dtype=bytes)
        # use view to cast to integer array
        # apply same transformation to reduce table size
        iseq = pseq.view('|S1').view(np.uint8) - 64
        iseq = np.clip(iseq, 1, 8)
        return iseq


    @staticmethod
    def _translate_into_kmer_indices(iseq: NDArray, k: int, table: NDArray) -> NDArray:
        """
        Translate the integer sequence into kmer indices using the kmer table.

        :param iseq: The integer array representing the sequence.
        :param k: The length of the kmers.
        :param table: The kmer table.
        :return: The array of kmer indices.
        """
        # use indexing to get unique numbers for imers
        ituple = tuple(iseq[i: -(k-i+1)] for i in range(k))
        imer_t = table[ituple]
        return imer_t


    def count(self, seq: str, k: int) -> dict[str, int]:
        """
        Count the occurrences of kmers in the given sequence.

        :param seq: The input sequence.
        :param k: The length of the kmers.
        :return: The dictionary of kmer counts.
        """
        # str to integer array
        iseq = self._integer_seq(seq)
        # grab the correct pre-computed table
        table = getattr(self, f't{k}')
        # translate to kmer indices
        tseq = self._translate_into_kmer_indices(iseq, k, table)
        # fast counting with bincount
        counts = np.bincount(tseq)
        # make a dict for results
        count_dict = dict(zip(getattr(self, f'k{k}'), counts))
        return count_dict


    @staticmethod
    def _expected_tetramer_frequencies(km: list[dict[str, int]]) -> dict[str, float]:
        """
        Calculate the expected tetramer frequencies.

        :param km: The list of kmer count dictionaries for di-, tri-, and tetramers.
        :return: The dictionary of expected tetramer frequencies.
        """
        tetra_exp = {}
        tetra = km[2]
        for tet in tetra:
            tetra_exp[tet] = (1.0 * km[1][tet[:3]] * km[1][tet[1:]] / km[0][tet[1:3]])
        return tetra_exp


    @staticmethod
    def _estimate_zscores(km: list[dict[str, int]], tetra_exp: dict[str, float]) -> dict[str, float]:
        """
        Estimate the z-scores for tetramers.

        :param km: list of kmer count dictionaries for di-, tri-, and tetramers.
        :param tetra_exp: dictionary of expected tetramer frequencies.
        :return: dictionary of tetramer z-scores.
        """
        tetra_sd = {}
        for tet, exp in list(tetra_exp.items()):
            den = km[0][tet[1:3]]
            tetra_sd[tet] = sqrt(exp * (den - km[1][tet[:3]]) * (den - km[1][tet[1:]]) / (den * den))

        tetra_z_num = np.array([km[2][tet] - exp for tet, exp in tetra_exp.items()])
        tetra_z_denom = np.array([tetra_sd[tet] for tet in tetra_exp.keys()])
        tetra_z_arr = np.divide(tetra_z_num, tetra_z_denom, where=tetra_z_denom > 0)
        tetra_z = dict(zip(tetra_exp.keys(), tetra_z_arr))
        return tetra_z


    def tetra_zscores(self, seq: str) -> dict[str, float]:
        """
        WRAPPER to calculate the z-scores for tetramers in the given sequence.

        :param seq: The input sequence.
        :return: The dictionary of tetramer z-scores.
        """
        # first count 2,3,4-mers
        kmers = [dict()] * 3
        kmers[0] = self.count(seq, 2)
        kmers[1] = self.count(seq, 3)
        kmers[2] = self.count(seq, 4)
        # calculate expected frequencies
        tetramer_exp = self._expected_tetramer_frequencies(kmers)
        tm_zscores = self._estimate_zscores(km=kmers, tetra_exp=tetramer_exp)
        return tm_zscores



class TetramerDist:

    def __init__(self):
        """
        This object takes sequence objects, not raw sequences.
        To avoid recomputing kmers.
        """
        self.kmc = KmerCounter()


    def euclidean_dist(self, seqo1: Any, seqo2: Any) -> float:
        """
        Calculate the Euclidean distance between two sequence objects.

        :param seqo1: first sequence object.
        :param seqo2: second sequence object.
        :return: Euclidean distance.
        """
        # check if kmers have already been counted
        t1 = getattr(seqo1, 'tmers', None)
        t2 = getattr(seqo2, 'tmers', None)
        # count them if not
        if not t1:
            seqo1.tmers = self.kmc.count(seqo1.seq, 4)
        if not t2:
            seqo2.tmers = self.kmc.count(seqo2.seq, 4)
        # grab the tetramers
        t1 = seqo1.tmers
        t2 = seqo2.tmers
        # intersection to get all 4mers present in both sequences
        tetramers = set(sorted(t1.keys())) & set(sorted(t2.keys()))
        c1 = np.array([t1[t] for t in tetramers])
        c2 = np.array([t2[t] for t in tetramers])
        # normalize
        n1 = c1 / np.sum(c1)
        n2 = c2 / np.sum(c2)
        # difference to calc Euclidean distance
        diff = n1 - n2
        euc = np.sqrt(np.sum(diff * diff))
        return euc




class IntraProb:

    def __init__(self):
        """
        This class calculates the intra-probability and threshold values.
        """
        self.mean = 0
        self.std = 0.01037897 / 2
        self.mean2 = 0.0676654
        self.std2 = 0.03419337
        # self.t = self.calc_threshold()
        self.t = 0.036  # hard-coded to avoid scipy dependency


    # def intra_prob(self, e: NDArray) -> float:
    #     """
    #     Calculate the intra-probability.
    #
    #     :param e: input value
    #     :return: intra-probability.
    #     """
    #     from scipy.stats import norm
    #     self.norm_intra = norm(self.mean, self.std)
    #     self.norm_inter = norm(self.mean2, self.std2)
    #     prob = self.norm_intra.pdf(e) / (self.norm_inter.pdf(e) + self.norm_intra.pdf(e))
    #     return prob
    #
    #
    # def calc_threshold(self) -> float:
    #     """
    #     Calculate the threshold value based on the empirical parameters.
    #
    #     :return: threshold value.
    #     """
    #     e = np.arange(0, 0.1, 0.001)
    #     ip = self.intra_prob(e)
    #     t = e[np.where(ip > 1e-10)][-1]
    #     return t


# instantiate to make available when imported
kmc = KmerCounter()
count_kmers = kmc.count                                 # basic counting function
tetramer_zscores = kmc.tetra_zscores                    # wrapper for tetramer z-scores

tdist = TetramerDist()
euclidean_dist = tdist.euclidean_dist                   # distance function for sequence OBJECTS
# pearson_dist = tdist.pearson_cor                      # distance function for sequence OBJECTS
euclidean_threshold = IntraProb().t                     # constant from empirical parameters














from minknow_api.manager import Manager

manager = Manager(host="localhost")

positions = list(manager.flow_cell_positions())
pos_dict = {pos.name: pos for pos in positions}
print(pos_dict)
target_device = pos_dict["MS00000"]
print(target_device)
device_connection = target_device.connect()
current_run = device_connection.protocol.get_current_protocol_run()
run_id = current_run.run_id
print(run_id)
out_path = current_run.output_path
print(out_path)




import logging

import numpy as np

from boss.paf import Paf, paf_dict_type



class AbundanceTracker:

    def __init__(self, contigs: dict):
        """
        Initialise the abundance tracker to count observed molecules

        :param contigs: Dictionary of contigs
        """
        self.total_reads = 0
        self.read_counts = {cname: 0 for cname in contigs.keys()}


    def _count_read_targets(self, paf_dict: paf_dict_type) -> None:
        """
        Count up the target sequences of the mappings in the paf_dict

        :param paf_dict: Dict of mappings
        :return:
        """
        # in case there was no mapped read in the batch
        if len(paf_dict) == 0:
            return

        for rid, rec in paf_dict.items():
            # select best mapper if there are multiple
            if len(rec) > 1:
                rec = Paf.choose_best_mapper(rec)
            # grab target of the mapping
            t = rec[0].tname
            try:
                self.read_counts[t] += 1
            except KeyError:
                pass


    def _report_proportions(self) -> None:
        """
        Calculate and report the proportions of reads per target

        :return:
        """
        props = {tname: (n / self.total_reads) for tname, n in self.read_counts.items()}
        logging.info("Counts and rel. proportions of observed reads:")
        for t in list(props.keys()):
            logging.info(f"{t}: {self.read_counts[t]} {np.round(props[t], 3)}")



    def update(self, n: int, paf_dict: paf_dict_type) -> None:
        """
        At each update we increment the observed read count,
        and count the targets of observed reads

        :param n: Total number of new reads
        :param paf_dict: Dict of mappings
        :return:
        """
        self.total_reads += n
        self._count_read_targets(paf_dict)
        self._report_proportions()







import logging

from boss.sampler import Sampler
from boss.batch import ReadCache
from boss.mapper import Mapper
from boss.paf import Paf, paf_dict_type
from boss.aeons.core import BossAeons
from boss.aeons.sequences import SequencePool
from boss.aeons.repeats import RepeatFilter


class BossAeonsSim(BossAeons):



    def init_sim(self) -> None:
        """
        Initialisation wrapper for simulated experiments

        :return:
        """
        # run the init from super
        self.init()

        # initialise the sampler
        self.sampler = Sampler(
            source=self.args.fq,
            maxbatch=self.args.maxb,
            batchsize=self.args.batchsize
        )
        # initialise pseudotiming object
        self.read_cache = ReadCache(batchsize=self.args.batchsize, dumptime=self.args.dumptime)
        # run initial assembly with starting data
        self._initial_asm()
        self.strat = {}  # empty strat to accept everything




    def _initial_asm(self):
        # load some initial batches
        init_pool = SequencePool(name="init_pool", out_dir=self.out_dir)
        for i in range(self.args.binit):
            read_sequences, _, _, _ = self.sampler.sample()
            init_pool.ingest(seqs=read_sequences)
        logging.info(f"total bases in pool: {init_pool.total_bases()}")
        # increment time after preloading
        self.read_cache.update_times_aeons(read_sequences=init_pool.seqdict(),
                                           reads_decision=init_pool.seqdict())
        # set the batch counter
        self.batch = self.sampler.fq_stream.batch
        # initialise a repeat filter from the raw initial reads
        if self.args.filter_repeats:
            self.repeat_filter = RepeatFilter(name=self.args.name, seqpool=init_pool)
        # run first asm
        logging.info("Running assembly of initial data..")
        init_contigs = init_pool.initial_asm_miniasm()
        self.pool.ingest(init_contigs)
        has_contig = self.pool.has_min_one_contig(min_contig_len=self.args.min_contig_len)
        ncontigs = len(self.pool.sequences)
        logging.info(f'initial contigs: {ncontigs}')
        if len(self.pool.sequences) == 0 or not has_contig:
            raise ValueError("No contigs of sufficient length, restart simulation with more data (binit)")
        self.pool.write_seq_dict(seq_dict=self.pool.seqdict(), file=self.pool.contig_fa)



    def make_decisions(
        self,
        paf_dict: paf_dict_type,
        read_sequences: dict,
        window: int = 100,
        mu: int = 400
    ) -> dict:
        """
        Decision function for simulations only.
        In real experiments readfish performs this functionality

        :param paf_dict: Paf dict of mapping runcated reads to contigs
        :param read_sequences: Dictionary of new read sequences
        :param window: Size of downsampling window
        :return: Dict of read sequences after decisions, i.e. rejected reads are truncated
        """
        # if nothing mapped, just return. Unmapped = accept
        if len(paf_dict.items()) == 0:
            logging.info("nothing mapped")
            self.reject_count = 0
            self.accept_count = 0
            self.unmapped_count = 0
            return read_sequences

        reads_decision = dict()
        reject_count = 0
        accept_count = 0
        unmapped_count = 0
        # loop over paf dictionary
        for record_id, record_list in paf_dict.items():
            # record_id, record_list = list(gaf_dict.items())[0]
            if len(record_list) > 1:
                # should not happen often since we filter secondary mappings
                rec = Paf.choose_best_mapper(record_list)[0]
            else:
                rec = record_list[0]

            # find the start and end position relative to the whole linearised genome
            if rec.rev:
                start_pos = rec.tend - 1
            else:
                start_pos = rec.tstart

            # index into strategy to find the decision
            try:
                decision = self.strat[str(rec.tname)][start_pos // window, rec.rev]
            except:
                # if we don't have a strategy yet accept all
                decision = 1

            # ACCEPT
            if decision:
                record_seq = read_sequences[rec.qname]
                accept_count += 1

            # REJECT
            else:
                record_seq = read_sequences[rec.qname][: mu]
                reject_count += 1

            # append the read's sequence to a new dictionary of the batch after decision-making
            reads_decision[rec.qname] = record_seq

        # all unmapped reads also need to be accepted, i.e. added back into the dict
        mapped_ids = set(reads_decision.keys())

        for read_id, seq in read_sequences.items():
            if read_id in mapped_ids:
                continue
            else:
                reads_decision[read_id] = seq
                unmapped_count += 1

        logging.info(f'decisions - rejecting: {reject_count} accepting: {accept_count} unmapped: {unmapped_count}')
        self.reject_count = reject_count
        self.accept_count = accept_count
        self.unmapped_count = unmapped_count
        return reads_decision





    def process_batch_aeons_sim(self) -> None:
        """
        Process new batch of simulated data. Get new reads, make decisions,
        update times and write to files before launching the common steps of
        processing new data

        :return:
        """
        read_sequences, _, _, _ = self.sampler.sample()
        # initialise a new mapper for the current contigs
        lm = Mapper(ref=self.pool.contig_fa, default=False)
        # map the truncated version of reads to the current contigs
        paf_trunc = lm.map_sequences(sequences=read_sequences, trunc=True)
        # Dict of processed reads, i.e. rejected sequences are truncated
        reads_decision = self.make_decisions(paf_dict=paf_trunc, read_sequences=read_sequences)
        # update read length dist
        self.rl_dist.update(read_lengths=self.sampler.fq_stream.read_lengths)
        # updates to the pseudotimes
        self.read_cache.update_times_aeons(read_sequences=read_sequences, reads_decision=reads_decision)
        self.read_cache.fill_cache(read_sequences=read_sequences, reads_decision=reads_decision)
        # run the rest of the steps to update
        self.update_wrapper(new_reads=reads_decision)








import logging
from collections import defaultdict
from io import StringIO
from copy import deepcopy

from boss.runs.core import BossRuns
from boss.sampler import Sampler
from boss.paf import Paf, PafLine, paf_dict_type
from boss.batch import ReadCache


class BossRunsSim(BossRuns):

    def init_sim(self):
        # run the init from super
        self.init()

        # initialise the sampler
        self.sampler = Sampler(
            source=self.args.fq,
            paf_full=self.args.paf_full,
            paf_trunc=self.args.paf_trunc,
            maxbatch=self.args.maxb,
            batchsize=self.args.batchsize
        )
        # initialise pseudotiming object
        self.read_cache = ReadCache(batchsize=self.args.batchsize, dumptime=self.args.dumptime)
        self.mu = 400



    def make_decisions(
        self,
        seqs: dict[str, str],
        paf_full: str,
        paf_trunc: str,
        window: int = 100
    ) -> tuple[paf_dict_type, dict[str, str], int, int, int, int]:
        """
        Make decisions about sampled data and build the paf_dict

        :param seqs: Dict of raw sequences
        :param paf_full: Raw output of mapping full-length reads
        :param paf_trunc: Raw output of mapping truncated reads
        :param window: downsampling size
        :return: pad_dict, and numbers of unmapped and rejected reads
        """
        # build a paf dict and either accept or reject reads
        paf_dict = defaultdict(list)
        mapped_reads = set()
        n_rejected = 0
        n_accepted = 0
        # return a dict of sequences after decision process
        reads_decision = deepcopy(seqs)

        # parse the raw paf data into dictionaries
        paf_dict_full = Paf.parse_PAF(StringIO(paf_full))
        paf_dict_trunc = Paf.parse_PAF(StringIO(paf_trunc))

        # loop over the mu-sized mappings to make decisions
        for rid, rlist in paf_dict_trunc.items():
            rec = Paf.choose_best_mapper(rlist)[0]
            mapped_reads.add(rid)
            # deal with strandedness
            if rec.rev:
                start_pos = rec.tend - 1
            else:
                start_pos = rec.tstart
            # actual decision look-up
            try:
                strat = self.contigs_filt[str(rec.tname)].strat
                decision = strat[start_pos // window, rec.rev]

            except (KeyError, IndexError):
                # in case the read maps to a chromosome that we don't have a strategy for
                # reject by default (can happen in testing or for references with short scaffolds)
                decision = 0

            if decision:
                # ACCEPT READ
                # grab the full-length version
                rec_full_list = paf_dict_full[str(rec.qname)]
                rec_full = Paf.choose_best_mapper(rec_full_list)[0]
                paf_dict[str(rec.qname)].append(rec_full)
                n_accepted += 1
            else:
                # REJECT READ
                paf_dict[str(rec.qname)].append(rec)
                n_rejected += 1
                reads_decision[rid] = reads_decision[rid][:self.mu]


        n_mapped = len(mapped_reads)
        n_unmapped = len(self.sampler.fq_stream.read_ids - mapped_reads)
        return paf_dict, reads_decision, n_mapped, n_unmapped, n_accepted, n_rejected



    def filter_paf_dict(self, paf_dict: dict[str, list[PafLine]]) -> dict[str, list[PafLine]]:
        """
        Filter the paf_dict down to only the reads that were accepted
        :param paf_dict:
        :return:
        """
        paf_dict_acc = {}
        for rid, reclist in paf_dict.items():
            rec = reclist[0]
            if rec.qlen != self.read_cache.mu:
                paf_dict_acc[rid] = reclist
        return paf_dict_acc



    def process_batch_runs_sim(self) -> None:
        """
        Process a new batch of simulations, only the first part is different to live runs
        The second part is run inside the update_wrapper() of the superclass
        :return:
        """
        # trigger the sampling of reads and their mappings
        read_seqs, read_quals, paf_f, paf_t = self.sampler.sample()
        # make decisions and generate the paf_dict
        paf_dict, reads_decision, n_mapped, n_unmapped, n_accepted, n_rejected = (
            self.make_decisions(seqs=read_seqs,
                                paf_full=paf_f,
                                paf_trunc=paf_t
                                )
        )
        logging.info(f"mapped {n_mapped}, not mapped {n_unmapped}")
        logging.info(f"accepted {n_accepted}, rejected {n_rejected}")
        # filter paf_dict to accepted reads only
        paf_dict_acc = self.filter_paf_dict(paf_dict=paf_dict)
        # update read length distribution
        self.rl_dist.update(read_lengths={n: r[0].qlen for n, r in paf_dict_acc.items()})
        # NOTE from here similar but not identical to live version
        # convert coverage counts to increment arrays
        increments = self.cc.convert_records(paf_dict=paf_dict, seqs=read_seqs, quals=read_quals)
        # effect the coverage increments for each contig
        self._effect_increments(increments=increments)
        # update the abundance tracker with new data
        self.tracker.update(n=n_accepted, paf_dict=paf_dict_acc)
        # update read starting position distribution
        self.read_starts.count_read_starts(paf_dict=paf_dict_acc)
        # update pseudotimes and check if it's time to write cache to file
        self.read_cache.update_times_runs(
            total_bases=self.sampler.fq_stream.total_bases,
            paf_dict=paf_dict,
            n_unmapped=n_unmapped,
            n_reject=n_rejected
        )
        self.read_cache.fill_cache(read_sequences=self.sampler.fq_stream.read_sequences, reads_decision=reads_decision)
        # update wrapper of superclass
        self.update_wrapper()













"""Run a targeted sequencing experiment on a device.

When given a :doc:`TOML <toml>` experiment configuration ``readfish targets`` will:
    #. Initialise a connection to the sequencing device
    #. Load the experiment configuration
    #. Initialise a connection to the Read Until API
    #. Initialise a connection to your chosen basecaller
    #. Initialise the read aligner

Then, during sequencing the start of each read is sampled.
These chunks of raw data are processed by the basecaller to produce FASTA, which is then aligned against the chosen reference genome.
The result of the alignment is used, along with the targets provided in the :doc:`TOML <toml>` file, to make a decision on each read.

In the new **experimental** Duplex mode, it is possible to override a decision for a read based on the action taken for a previous read.
This is done by passing `--chemistry` and setting either duplex, or duplex simple.
`duplex_simple` accepts a read if the previous channels read was stop receiving, `duplex` checks that the previous reads alignment was on the same contig and opposite strand.
The default chemistry is simplex.

Running this should result in a very short (<1kb, ideally 400-600 bases) unblock peak at the start of a read length histogram and longer sequenced reads.

Example run command::

   readfish targets --device X3 \\
           --experiment-name "test" \\
           --toml my_exp.toml \\
           --log-file rf.log \\
           --debug-log chunks.tsv

Example experimental duplex command::

    readfish targets --device X3 \\
           --experiment-name "test" \\
           --toml my_exp.toml \\
           --log-file rf.log \\
           --debug-log chunks.tsv
           --chemistry duplex

In the debug_log chunks.tsv file, if this argument is passed, each line represents detailed information about a batch of
read signal that has been processed in an iteration.

The format of each line is as follows:
loop_counter  number_reads  read_id  channel  read_number  seq_length  seen_count
decision  action  condition  barcode  previous_action  timestamp  action_overridden

+-------------------+-------------+---------------------------------------------------------------------------------------------------------------+
| Parameter         | Type        | Description                                                                                                   |
+===================+=============+===============================================================================================================+
| loop_counter      | int         | The iteration number for the loop.                                                                            |
+-------------------+-------------+---------------------------------------------------------------------------------------------------------------+
| number_reads      | int         | The number of reads processed in this iteration.                                                              |
+-------------------+-------------+---------------------------------------------------------------------------------------------------------------+
| read_id           | str         | UUID4 string representing the reads unique read_id.                                                           |
+-------------------+-------------+---------------------------------------------------------------------------------------------------------------+
| channel           | int         | Channel number the read is being sequenced on.                                                                |
+-------------------+-------------+---------------------------------------------------------------------------------------------------------------+
| read_number       | int         | The number this read is in the sequencing run as a whole.                                                     |
+-------------------+-------------+---------------------------------------------------------------------------------------------------------------+
| seq_length        | int         | Length of the base-called signal chunk (includes any previous chunks).                                        |
+-------------------+-------------+---------------------------------------------------------------------------------------------------------------+
| seen_count        | int         | Number of times this read has been seen in previous iterations.                                               |
+-------------------+-------------+---------------------------------------------------------------------------------------------------------------+
| decision          | str         | The name of the Decision variant taken for this read, see :ref:`regions-sub-tables` for values.               |
+-------------------+-------------+---------------------------------------------------------------------------------------------------------------+
| action            | str         | The name of the Action variant sent to the sequencer for this read, see :ref:`regions-sub-tables` for values. |
+-------------------+-------------+---------------------------------------------------------------------------------------------------------------+
| condition         | str         | Name of the Condition that the read has been addressed with.                                                  |
+-------------------+-------------+---------------------------------------------------------------------------------------------------------------+
| barcode           | str or None | The name of the Barcode for this read if present, otherwise None.                                             |
+-------------------+-------------+---------------------------------------------------------------------------------------------------------------+
| previous_action   | str or None | Name of the last Action taken for a read sequenced by this channel or None if first read on a chann           |
+-------------------+-------------+---------------------------------------------------------------------------------------------------------------+
| action_overridden | bool        | Indicates if the action has been overridden. Currently, actions are always overridden to be `stop_receiving`. |
+-------------------+-------------+---------------------------------------------------------------------------------------------------------------+
| timestamp         | float       | Current time as given by the time module in seconds.                                                          |
+-------------------+-------------+---------------------------------------------------------------------------------------------------------------+

Actions being overridden occurs when the readfish run is a dry run and the action is unblock, or when the read is the first read seen for a channel by readfish.
This prevents trying to unblock reads of unknown length.


"""

# Core imports
from __future__ import annotations
import argparse
import traceback
from packaging.version import Version
import logging
import time
from timeit import default_timer as timer
from pathlib import Path
from typing import Any

# Third party imports
from readfish.read_until.read_cache import AccumulatingCache
from readfish.read_until import ReadUntilClient
from minknow_api import protocol_service

# Library
from readfish._cli_args import DEVICE_BASE_ARGS, Chemistry
from readfish._read_until_client import RUClient
from readfish._config import Action, Conf, make_decision, _Condition
from readfish._statistics import ReadfishStatistics
from readfish.__about__ import __version__
from readfish._compatibility import (
    _get_minknow_version,
    check_compatibility,
    MINKNOW_COMPATIBILITY_RANGE,
    DIRECTION,
)
from readfish._utils import (
    get_device,
    send_message,
    ChunkTracker,
    Severity,
)
from readfish.plugins.abc import AlignerABC, CallerABC
from readfish.plugins.utils import (
    Decision,
    PreviouslySentActionTracker,
    Result,
    DuplexTracker,
    Strand,
)


# TODO custom imports
from boss.dynamic_readfish import BossBits, get_args
import sys



_help = "Run targeted sequencing"
_cli = DEVICE_BASE_ARGS + (
    (
        "--toml",
        dict(
            metavar="TOML",
            required=True,
            help="TOML file specifying experimental parameters",
        ),
    ),
    (
        "--no-debug-log",
        dict(
            help="Disable debug output of information about chunks seen into a .tsv formatted log. Default enabled.",
            action="store_false",
            dest="debug_log",
        ),
    ),
    (
        "--padding",
        dict(
            help="Number of bases to pad the target sequences with",
            default=0,
            type=int,
            metavar="PADDING",
        ),
    ),
)
# When sequencing in duplex mode, overriding a decided `Action` on a currently sequenced molecule
# is not allowed if the previous molecules decision was one of these.
DISALLOWED_DUPLEX_DECISIONS = {Decision.first_read_override, Decision.duplex_override}


class Analysis:
    """
    Analysis class where the read until magic happens. Comprises of one run
    function that is run threaded in the run function at the base of this file.
    Arguments listed in the __init__ docs.

    :param client: An instance of the ReadUntilClient object.
    :param conf: An instance of the Conf object.
    :param logger: The command level logger for this module.
    :param debug_log: Whether to output the Debug Log. log Name is generated.
    :param throttle: The time interval (seconds) between requests to the ReadUntilClient.
    :param unblock_duration: Time, in seconds, to apply unblock voltage.
    :param dry_run: If True unblocks are replaced with `stop_receiving` commands.
    :param toml: The path to the toml file containing experiment conf. Used as the path for checking if the TOML needs reloading.
    :param chemistry: Instance of Chemistry Enum, representing the chemistry of the run (Simplex/Duplex). Used for
        decision making on strands that may be part of a duplex pair.
    """

    def __init__(
        self,
        client: ReadUntilClient,
        conf: Conf,
        logger: logging.Logger,
        debug_log: bool,
        throttle: float,
        unblock_duration: float,
        dry_run: bool,
        toml: str,
        chemistry: Chemistry,
    ):
        self.client = client
        self.conf = conf
        self.logger = logger
        self.debug_log = debug_log
        self.throttle = throttle
        self.unblock_duration = unblock_duration
        self.dry_run = dry_run
        self.live_toml = Path(f"{toml}_live").resolve()
        self.run_information = self.client.connection.protocol.get_run_info()
        self.chemistry = chemistry
        # Generate a run specific read log
        read_log_name = (
            f"{self.run_information.run_id}_readfish.tsv" if debug_log else None
        )

        self.logger.info("Fetching Run Configuration")
        self.break_reads_after_seconds = (
            self.client.connection.analysis_configuration.get_analysis_configuration().read_detection.break_reads_after_seconds.value
        )
        self.sample_rate = self.client.connection.device.get_sample_rate().sample_rate
        self.logger.info("Run Configuration Received")
        self.logger.info(f"run_id={self.run_information.run_id}")
        self.logger.info(f"break_reads_after_seconds={self.break_reads_after_seconds}")
        self.logger.info(f"sample_rate={self.sample_rate}")
        # Create our statistics tracker
        self.loop_statistics = ReadfishStatistics(
            read_log_name, self.break_reads_after_seconds
        )
        logger.info("Initialising Caller")
        self.caller: CallerABC = self.conf.caller_settings.load_object(
            "Caller", run_information=self.run_information, sample_rate=self.sample_rate
        )
        logger.info("Caller initialised")
        caller_description = self.caller.describe()
        self.logger.info(caller_description)
        send_message(self.client.connection, caller_description, Severity.INFO)
        logger.info("Initialising Aligner")
        self.mapper: AlignerABC = self.conf.mapper_settings.load_object("Aligner")
        self.logger.info("Aligner initialised")
        # count how often a read is seen
        self.chunk_tracker = ChunkTracker(self.client.channel_count)

        # This is an object to keep track of the last action sent to the client for each channel
        self.previous_action_tracker = PreviouslySentActionTracker()
        # Keep track of previous alignments
        self.duplex_tracker = DuplexTracker()

        # We assume that sequencing is already running.
        # If the run is not in sequencing phase when the read until loop starts will
        # be set to false and the first read seen can be unblocked.
        self.readfish_started_during_sequencing = True

        # This is a flag to prevent repeated logging of the same message
        self.log_once_in_loop = True

    @property
    def wait_for_sequencing(self) -> bool:
        """
        Wait for minKNOW to report PHASE_SEQUENCING before starting readfish tight loop.
        The check occurs in out RUClient wrapper.

        :return: True if we are waiting for PHASE_SEQUENCING, False otherwise

        """
        if self.client.wait_for_sequencing_to_start:
            if self.log_once_in_loop:
                self.logger.info(
                    f"MinKNOW is reporting {protocol_service.ProtocolPhase.Name(self.client.current_protocol_phase)}, waiting for PHASE_SEQUENCING to begin."
                )
                self.log_once_in_loop = not self.log_once_in_loop
            self.readfish_started_during_sequencing = False  # We are not in sequencing phase, so we can unblock the first read we see as we will be sequencing it from the start
            return True
        return False

    def reload_toml(self, last_toml_mtime: float) -> float:
        """
        Reload the toml to refresh the conf with any updates.
        Reloading is determined by checking the modified time of the toml file.
        If it is more recent, reload the conf.

        :param last_live_mtime: The last modified time for the toml file.

        :return: The last modified time for the toml file, updated if changed.
        """
        if (
            self.live_toml.is_file()
            and self.live_toml.stat().st_mtime > last_toml_mtime
        ):
            try:
                self.conf = Conf.from_file(
                    self.live_toml, self.client.channel_count, self.logger
                )
            # FIXME: Broad exception
            except Exception as e:
                if hasattr(e, "exceptions"):
                    self.logger.error(getattr(e, "exceptions"))
                    self.logger.error(traceback.format_exc())
            last_toml_mtime = self.live_toml.stat().st_mtime
        return last_toml_mtime

    def check_override_action(
        self,
        control: bool,
        action: Action,
        result: Result,
        seen_count: int,
        condition: _Condition,
        stop_receiving_action_list: list[tuple[int, int]],
        unblock_batch_action_list: list[tuple[int, int]],
    ) -> tuple[Action, Action, bool, str | None]:
        """
        Check the chosen Action and amend it based on conditional checks.
        The action lists are appended to in place, so no return is required.

        Checks include:
            1. If the read is in a control region, the action is always stop_receiving.
            1. If the read is below the minimum chunks, use value in toml or default to proceed
            1. If the read is above the maximum chunks, use value in toml or default unblock--throttle
            1. First read seen for channel and readfish started during sequencing, override to stop_receiving
            1. If action is unblock and we are dry-running, override to stop_receiving
            1. If we are running in duplex chemistry, check the previous reads final decision and Action, and potentially sequence
                the current read, instead of unblocking it.

        :param control: Indicates read from a channel in a control region
        :param action: What action was decided for this read before any meddling
        :param result: Information about the current read.
        :param seen_count: Number of times other chunks from the read have been observed.
        :param condition: The set of conditions for deciding the action.
        :param stop_receiving_action_list: List to append channels and read numbers for which 'stop receiving' action is decided.
        :param unblock_batch_action_list: List to append channels, read numbers, and read IDs for which 'unblock' action is decided.

        :return: A tuple containing the previous action taken for this read,
          boolean indicating if the action was overridden, and the name of the action overridden too.

        """

        # Easy dub
        if control:
            action = Action.stop_receiving
        else:
            # TODO: Document the less than logic here
            below_min_chunks = seen_count < condition.min_chunks
            above_max_chunks = seen_count > condition.max_chunks

            # TODO: This will also factor into the precedence and documentation
            # If we have seen this read more than the max chunks and want to
            #   evaluate it again (Action.proceed) then we will overrule that
            #   action using the above_max_chunks_action, unblock by default
            if above_max_chunks and action is Action.proceed:
                action = condition.above_max_chunks
                result.decision = Decision.above_max_chunks

            # If we are below min chunks and we get an action that is not PROCEED
            #   then we will overrule that action using the below_min_chunks_action
            #   which by default is proceed.
            if below_min_chunks and action is not Action.proceed:
                action = condition.below_min_chunks
                result.decision = Decision.below_min_chunks

        # previous_action will be None if the read has not been seen before.
        previous_action = self.previous_action_tracker.get_action(result.channel)
        action_overridden = False
        # If --duplex flag override decisions made based on the strand and contig alignment of the previous read.
        # Unfinished bruv
        if (
            self.chemistry is Chemistry.DUPLEX
            # Easy checks first, so wdon't do more complex processing unless we have to
            and action == Action.unblock
            and previous_action is Action.stop_receiving
        ):
            # Check if we think this read is possibly duplex
            possible_duplex = any(
                self.duplex_tracker.possible_duplex(
                    result.channel, result.read_id, al.ctg, al.strand
                )
                for al in result.alignment_data
            )
            # Check the previous decision for this channel was not already an override
            previous_decision_allowed = (
                self.duplex_tracker.get_previous_decision(result.channel)
                not in DISALLOWED_DUPLEX_DECISIONS
            )
            if possible_duplex and previous_decision_allowed:
                self.logger.debug(
                    f"Overriding read {result.read_id} as it is possibly second half of a duplex"
                    f"- previous read action {previous_action}, current_action: {action},"
                    f" previous_decision: {self.duplex_tracker.get_previous_decision(result.channel)}"
                )
                action_overridden = True
                result.decision = Decision.duplex_override
                action = Action.stop_receiving
        # Duplex
        elif (
            self.chemistry is Chemistry.DUPLEX_SIMPLE
            and previous_action is Action.stop_receiving
            and action is Action.unblock
        ):
            previous_decision_allowed = (
                self.duplex_tracker.get_previous_decision(result.channel)
                not in DISALLOWED_DUPLEX_DECISIONS
            )
            if previous_decision_allowed:
                self.logger.debug(
                    f"Overriding to duplex - previous read action {previous_action}, current_action: {action},"
                    f" previous_decision: {self.duplex_tracker.get_previous_decision(result.channel)}"
                )
                action = Action.stop_receiving
                action_overridden = True
                result.decision = Decision.duplex_override

        # Override to stop receiving if this is the first read ona channel and we started mid sequencing
        if previous_action is None and self.readfish_started_during_sequencing:
            self.logger.debug(
                f"This is the first suitable read chunk from channel {result.channel}. Translocated read length unknown, sequencing."
            )
            action_overridden = True
            result.decision = Decision.first_read_override
            action = Action.stop_receiving

        if action is Action.stop_receiving:
            stop_receiving_action_list.append((result.channel, result.read_id))

        elif action is Action.unblock:
            if self.dry_run:
                # Log an 'unblock' action to previous action, but send a 'stop receiving' to prevent further read processing.
                action_overridden = True
                stop_receiving_action_list.append((result.channel, result.read_id))
            else:
                unblock_batch_action_list.append((result.channel, result.read_id))

        # If we have made a final decision for this read and we shouldn't see it again!
        if action is Action.unblock or action is Action.stop_receiving:
            # Add decided Action
            self.previous_action_tracker.add_action(result.channel, action)
            # Add duplex based tracking if we are in duplex mode
            if self.chemistry is Chemistry.DUPLEX_SIMPLE:
                self.duplex_tracker.set_decision(result.channel, result.decision)
            elif self.chemistry is Chemistry.DUPLEX:
                self.duplex_tracker.set_decision(result.channel, result.decision)
                self.duplex_tracker.set_alignments(
                    result.channel,
                    [(al.ctg, Strand(al.strand)) for al in result.alignment_data],
                )

        return (
            action,
            previous_action,
            action_overridden,
            action.name if action_overridden else None,
        )

    def run(self):
        """Run the read until loop, in one continuous while loop."""

        # TODO: Swap this for a CSV record later
        self.conf.write_channels_toml(self.client.mk_run_dir)

        # TODO: This could still be passed through to the basecaller to prevent
        #       rebasecalling data that is already being unblocked or sequenced
        loop_counter = 0

        last_live_toml_mtime = 0
        self.logger.info("Starting main loop")
        self.logger.info("Generating aligner description, if possible...")
        mapper_description = self.mapper.describe(self.conf.regions, self.conf.barcodes)
        self.logger.info(mapper_description)
        send_message(self.client.connection, mapper_description, Severity.INFO)

        # TODO initialise parts needed for dynamic decisions
        boss = BossBits(conf=self.conf, logger=self.logger, mapper=self.mapper)


        while self.client.is_sequencing:
            t0 = timer()
            # Check if we have started readfish before PHASE_SEQUENCING,
            if self.wait_for_sequencing:
                time.sleep(self.throttle)
                continue
            # Set back to true for when we re-enter a non sequencing phase
            self.log_once_in_loop = True

            if self.readfish_started_during_sequencing and loop_counter == 0:
                self.logger.info(
                    "readfish started in PHASE_SEQUENCING. Fully sequencing first read from each channel."
                )
            if not self.mapper.initialised:
                self.logger.warning(
                    "readfish main loop started but mapper is not initialised. Please check your aligners plugin documentation."
                    "If you are using mappy or mappy-rs this is definitely an error. Please open an issue here - "
                    "https://github.com/LooseLab/readfish/issues"
                )
                time.sleep(self.throttle)
                continue

            # TODO load updated decision masks and contigs
            reload_mapper, mapper = boss.reload(conf=self.conf)
            if reload_mapper:
                self.mapper = mapper  # replace the mapper with new one

            last_live_toml_mtime = self.reload_toml(last_live_toml_mtime)
            ########### Main Loop ###########
            loop_counter += 1
            number_reads = 0
            unblock_batch_action_list = []
            stop_receiving_action_list = []

            chunks = self.client.get_read_chunks(self.client.channel_count, last=True)
            calls = self.caller.basecall(
                chunks, self.client.signal_dtype, self.client.calibration_values
            )
            aligns = self.mapper.map_reads(calls)

            #######################################################################
            for result in aligns:
                number_reads += 1
                control, condition = self.conf.get_conditions(
                    result.channel, result.barcode
                )
                # result.decision = make_decision(self.conf, result)
                # TODO decisions
                result.decision = boss.make_decision_boss(self.conf, result)
                action = condition.get_action(result.decision)
                seen_count = self.chunk_tracker.seen(result.channel, result.read_id)
                #  Check if there any conditions that override the action chose, exceed_max_chunks etc...
                (
                    action,
                    previous_action,
                    action_overridden,
                    overridden_action_name,
                ) = self.check_override_action(
                    control,
                    action,
                    result,
                    seen_count,
                    condition,
                    stop_receiving_action_list,
                    unblock_batch_action_list,
                )
                self.loop_statistics.log_read(
                    client_iteration=loop_counter,
                    read_in_loop=number_reads,
                    read_id=result.read_id,
                    channel=result.channel,
                    seq_len=len(result.seq),
                    counter=seen_count,
                    mode=result.decision.name,
                    decision=action.name,
                    condition=condition.name,
                    barcode=result.barcode,
                    previous_action=(
                        previous_action.name
                        if previous_action is not None
                        else previous_action
                    ),
                    action_overridden=action_overridden,
                    timestamp=time.time(),
                    # Anything below here is not included in the Debug log
                    region_name=(
                        _region.name
                        if (_region := self.conf.get_region(result.channel))
                        else "flowcell"
                    ),
                    overridden_action_name=overridden_action_name,
                )

            #######################################################################
            # Compile actions to be sent
            self.client.unblock_read_batch(
                unblock_batch_action_list, duration=self.unblock_duration
            )
            self.client.stop_receiving_batch(stop_receiving_action_list)

            t1 = timer()
            if number_reads > 0:
                self.loop_statistics.add_batch_performance(
                    number_of_reads=number_reads, batch_time=t1 - t0
                )
                self.logger.info(self.loop_statistics.get_batch_performance())

            # limit the rate at which we make requests
            if t0 + self.throttle > t1:
                time.sleep(self.throttle + t0 - t1)
        else:
            send_message(
                self.client.connection,
                "Readfish client stopped.",
                Severity.WARN,
            )
            self.caller.disconnect()
            self.mapper.disconnect()
            self.logger.info("Finished analysis of reads as client stopped.")


# TODO changing typehints here to allow pass-through and testing
def run(
    parser: argparse.ArgumentParser, args: argparse.Namespace, extras: list[Any]
) -> int | tuple[Conf, logging.Logger]:
    """Run function for targets.py

    Imported in `_cli_base.py`.
    Sets up the read until client and starts the analysis thread above.

    :param parser: Argparse onject - unused but must be taken due as may be needed
    :param args: The arguments passed to ArgParse
    :param extras: Extra stuff, I guess

    :returns: An exit code integer, 0 for success
    """
    # Setup logger used in this entry point, this one should be passed through
    logger = logging.getLogger(f"readfish.{args.command}")

    # Check MinKNOW version

    minknow_version = _get_minknow_version(host=args.host, port=args.port)
    if (
        action := check_compatibility(minknow_version, MINKNOW_COMPATIBILITY_RANGE)
    ) in (
        DIRECTION.UPGRADE,
        DIRECTION.DOWNGRADE,
    ):
        lower_bound, upper_bound = MINKNOW_COMPATIBILITY_RANGE
        logger.warning(
            f"""This readfish version ({__version__}) is tested for compatibility with MinKNOW v{lower_bound} to v{upper_bound}.
This version of minknow is {minknow_version}.
If readfish fails please try to {action.value} readfish.
If there isn't a newer version of readfish and readfish is failing, please open an issue:
    https://github.com/LooseLab/readfish/issues"""
        )

    if minknow_version < Version("6.0.0"):
        logger.critical(
            f"This version of readfish ({__version__}) is not compatible with less than MinKNOW 6.X.X, you downgrade to at least readfish 2024.2.0"
            f"This won't work, exiting..."
        )
        raise SystemExit(1)

    # Fetch sequencing device
    position = get_device(args.device, host=args.host, port=args.port)

    # Create a read until client
    read_until_client = RUClient(
        mk_host=position.host,
        mk_port=position.description.rpc_ports.secure,
        filter_strands=True,
        cache_type=AccumulatingCache,
        timeout=args.wait_for_ready,
        prefilter_classes={
            "strand",
            "strand2",
            "short_strand",
            "adapter",
            "unknown_positive",
        },
    )

    # Load TOML configuration
    conf = Conf.from_file(args.toml, read_until_client.channel_count, logger=logger)

    # Set the padding if it is specified.
    if padding := getattr(args, "padding", None):
        for region in conf.regions:
            region.targets.padding = padding
        for barcode in conf.barcodes:
            conf.barcodes[barcode].targets.padding = padding
    logger.info(conf.describe_experiment())

    send_message(
        read_until_client.connection,
        f"'readfish {args.command}' connected to this device.",
        Severity.WARN,
    )

    # TODO for testing purposes
    if hasattr(args, 'return_conf'):
        return conf, logger

    # start the client running
    read_until_client.run(
        first_channel=1,
        last_channel=read_until_client.channel_count,
        max_unblock_read_length_seconds=args.max_unblock_read_length_seconds,
        accepted_first_chunk_classifications=[
            "strand",
            "strand2",
            "short_strand",
            "adapter",
            "unknown_positive",
        ],
    )

    worker = Analysis(
        read_until_client,
        conf=conf,
        logger=logger,
        debug_log=args.debug_log,
        unblock_duration=args.unblock_duration,
        throttle=args.throttle,
        dry_run=args.dry_run,
        toml=args.toml,
        chemistry=Chemistry(args.chemistry),
    )

    # begin readfish function
    try:
        worker.run()
    except KeyboardInterrupt:
        logger.info("Keyboard interrupt received, stopping readfish.")
        pass
    finally:
        read_until_client.reset()

    send_message(
        read_until_client.connection,
        "Readfish disconnected from this device. Sequencing will proceed normally.",
        Severity.WARN,
    )
    return 0



def main(arg_list: list = None) -> None:
    """
    This is equivalent to launching entry-point in readfish from _cli_base.py
    in readfish main() of _cli_base.py would launch the entry-point,
    whereas here we grab the arguments and then launch a modified run function
    :param arg_list: List to pass in arguments for debugging
    :return:
    """
    BossBits.gen_dummy_idx()
    # get arguments
    if not arg_list:
        arg_list = sys.argv[1:]
    parser, args = get_args(arg_list)
    # launch entry-point
    run(parser=parser, args=args, extras=[])


if __name__ == "__main__":
    main()



from io import StringIO
from concurrent.futures import ThreadPoolExecutor as TPE
from pathlib import Path
import logging

import mappy

from .paf import Paf, paf_dict_type



class Indexer:

    def __init__(self, fasta: str, mmi: str, t: int = 4):
        """
        Initialize simple indexing wrapper around mm2

        :param fasta: The path to the input FASTA file.
        :param mmi: The path to the output .mmi index file.
        :param t: The number of threads to use for indexing.
        """
        self.aligner = mappy.Aligner(fn_idx_in=fasta, fn_idx_out=mmi, preset="map-ont", n_threads=t)



class Mapper:

    def __init__(self, ref: str, mu: int = 400, workers: int = 4, default: bool = True):
        """
        Initialize a Mapper object; wrapper for minimap2's mappy implementation
        For the default case of mapping against some linear references

        :param ref: The path to the reference FASTA file.
        :param mu: Length of anchor bases for simulations, defaults to 400.
        :param workers: The number of worker threads to use for mapping
        :param default: Whether to use the default mappy.Aligner configuration, defaults to True.
        """
        self.mu = mu
        self.workers = workers
        # check that the given reference exists
        if not Path(ref).is_file():
            raise FileNotFoundError("Given reference file does not exist")

        if default:
            self.aligner = mappy.Aligner(fn_idx_in=ref, preset="map-ont")
        else:
            self.aligner = mappy.Aligner(fn_idx_in=ref, fn_idx_out=f'{ref}.mmi', preset="map-ont",
                                         k=13, w=5, min_cnt=2, min_chain_score=20)



    def map_sequences(self, sequences: dict[str, str], trunc: bool = False) -> paf_dict_type:
        """
        Map sequences to the reference and return the mapping results.

        :param sequences: A dictionary of read_id: sequence items.
        :param trunc: Switch to perform mapping of mu-sized fragments for aeons simulations
        :return: A dictionary of read_id: list(PafLine) objects.
        """
        # truncation option for aeons simulations
        if trunc:
            sequences = {rid: seq[: self.mu] for rid, seq in sequences.items()}
        paf_raw = self._mappy_batch(sequences=sequences)
        paf_dict = Paf.parse_PAF(StringIO(paf_raw), min_len=int(self.mu / 2))
        return paf_dict



    def _mappy_batch(self, sequences: dict[str, str], out: str = None, log: bool = True) -> str:
        """
        Map a full batch of reads and return the PAF formatted mapping hits.

        :param sequences: A dictionary of read_id: sequence items.
        :param out: The name of the file to output mappings, defaults to None.
        :param log: Whether to log the number of mapped and unmapped queries, defaults to True.
        :return: The PAF formatted mapping hits of all mapped input reads.
        """
        # container to hold the hits from all reads
        batch_alignments = []
        unmapped_count = 0
        mapped_count = 0
        # loop over all sequences and map them one by one
        with TPE(max_workers=self.workers) as executor:
            results = executor.map(self._map_query, sequences.items())

        for result in results:
            res = '\n'.join(result)
            # prevent appending an empty list if the read was not mapped
            if len(res) > 0:
                batch_alignments.append(res)
                mapped_count += 1
            else:
                unmapped_count += 1

        # transform to a single string
        alignments = '\n'.join(batch_alignments)

        # tmp write to file too
        if out:
            with open(out, 'w') as lm_out:
                lm_out.write(alignments)

        # counting the number of mapped and unmapped fragments
        if log:
            logging.info(f"MAPPY: mapped queries: {mapped_count}, unmapped queries: {unmapped_count} ")
        self.mapped_count = mapped_count
        self.unmapped_count = unmapped_count
        return alignments



    def _map_query(self, query: tuple[str, str]) -> list[str]:
        """
        Map a query and return the mapped query as a list of PAF formatted mapping hits.

        :param query: A key-value pair of read_id and query.
        :return: A list of PAF formatted mapping hits.
        """
        results = []
        read_id, seq = query

        thr_buf = mappy.ThreadBuffer()
        # For each alignment to be mapped against, returns a PAF format line
        for hit in self.aligner.map(seq, buf=thr_buf):
            # if hit.is_primary:
            results.append(f"{read_id}\t{len(seq)}\t{hit}")
        return results







def add(queue, a, b):
    ret = a + b
    queue.put(ret)


def subtract(queue, a, b):
    ret = a - b
    queue.put(ret)


def divide(queue, a, b):
    if b == 0:
        raise ValueError("Cannot divide by zero")
    ret = a / b
    queue.put(ret)


def multiply(queue, a, b):
    ret = a * b
    queue.put(ret)


import logging
import re
import os
from pathlib import Path

import mappy
import numpy as np

from boss.utils import empty_file, execute, random_id
from boss.paf import paf_dict_type



class FastqBatch:

    def __init__(self, fq_files: list[str | Path], channels: set = None):
        """
        Initialise a new batch of sequencing reads using their filepaths

        :param fq_files: Paths of new files
        :param channels: Set of channels of the BOSS region
        """
        self.fq_files = fq_files
        self.channels = channels
        # attributes to be filled
        self.read_sequences = {}
        self.read_qualities = {}
        self.read_ids = {}
        self.read_lengths = {}
        self.total_bases = 0
        self._read_batch()



    def _read_batch(self) -> None:
        """
        Read sequencing data from all new files

        :return:
        """
        read_sequences = {}
        read_qualities = {}
        for fq in self.fq_files:
            rseq, rqual = self._read_single_batch(fastq_file=fq)
            read_sequences.update(rseq)
            read_qualities.update(rqual)
        # set attributes of the batch
        self.read_sequences = read_sequences
        self.read_qualities = read_qualities
        self.read_ids = set(read_sequences.keys())
        self.read_lengths = {rid: len(seq) for rid, seq in read_sequences.items()}
        self.total_bases = np.sum(list(self.read_lengths.values()))
        logging.info(f'total new reads: {len(read_sequences)}')



    def _read_single_batch(self, fastq_file: str) -> tuple[dict[str, str], dict[str, str]]:
        """
        Get the reads from a single fq file and put into dictionary

        :param fastq_file: A new fastq files
        :return: Dictionary of new sequences as header, sequence pairs
        """
        logging.info(f"Reading file: {fastq_file}")
        read_sequences = {}
        read_qualities = {}
        # to make sure it's a string
        if not isinstance(fastq_file, str):
            raise TypeError('Fastq file must be a string')
        # loop over all reads in the fastq file
        # if we consider all channels
        if not self.channels:
            for name, seq, qual, desc in mappy.fastx_read(fastq_file, read_comment=True):
                read_sequences[str(name)] = seq
                read_qualities[str(name)] = qual
        else:
            # consider source channel
            for name, seq, qual, desc in mappy.fastx_read(fastq_file, read_comment=True):
                try:
                    # regex to get the channel number from the header
                    # \s=whitespace followed by 'ch=' and then any amount of numeric characters
                    curr_channel = re.search("\sch=[0-9]*", desc).group()
                    ch_num = int(curr_channel.split('=')[1])
                except AttributeError:
                    # if the pattern is not in the header, skip the read
                    logging.info("ch= not found in header of fastq read")
                    continue
                # check if read comes from a BOSS channel
                if ch_num in self.channels:
                    read_sequences[str(name)] = seq
                    read_qualities[str(name)] = qual
        return read_sequences, read_qualities



class ReadCache:

    def __init__(self, batchsize: int, dumptime: int, alpha: int = 300, rho: int = 300, mu: int = 400):
        """
        Object for simulations only. Collects the pseudotime of different conditions and
        the reads seen by each condition. Then writes the observed reads to file in
        cumulative batches of data for downstream analysis. I.e. the files produced by this
        object can be used as timepoints in downstream analyses

        :param batchsize: How many reads are sampled in each batch
        :param dumptime: Time interval of writing data to file
        :param alpha: Acquisition time of model
        :param rho: Rejection time cost in model
        :param mu: Length of initial bit in model
        """
        self.alpha, self.rho, self.mu = alpha, rho, mu
        self.batchsize = batchsize
        # pseudotime counter
        self.time_boss = 0
        self.time_control = 0
        # for writing reads to file
        self.cache_control = dict()
        self.cache_boss = dict()
        # after how much time should sequences be written to file
        # dump time is incremented every time a batch is written, which happens once that is overcome
        self.dumptime = dumptime
        self.dump_n_control = 1
        self.dump_n_boss = 1
        # for storing the batches of reads for snakemake analysis
        if not os.path.exists('./00_reads'):
            os.mkdir('./00_reads')
        empty_file(f'00_reads/control_0.fa')
        empty_file(f'00_reads/boss_0.fa')




    def update_times_runs(self, total_bases: int, paf_dict: paf_dict_type, n_unmapped: int, n_reject: int) -> None:
        """
        Increment the pseudotime for control and boss regions on flowcell
        depending on the observed data and decisions made on them

        :param total_bases: Total observed bases in batch
        :param paf_dict: Dict of mappings
        :param n_unmapped: Number of unmapped reads
        :param n_reject: Number of rejected reads
        :return:
        """
        # Control: total number of bases + alpha for each read
        self.time_control += total_bases
        self.time_control += (self.batchsize * self.alpha)
        # BR: all accepted bases and ((mu + rho) * number of unmapped/rejected reads) + (alpha * batch_size)
        bases_br = np.sum([r[0].qlen for r in paf_dict.values()])
        self.time_boss += bases_br
        self.time_boss += (n_unmapped * (self.mu + self.rho))
        self.time_boss += (n_reject * (self.mu + self.rho))
        self.time_boss += (self.batchsize * self.alpha)
        logging.info(f"time control: {self.time_control}")
        logging.info(f"time boss-runs: {self.time_boss}")



    def update_times_aeons(self, read_sequences: dict[str, str], reads_decision: dict[str, str]) -> None:
        """
        Increment pseudotime for control and boss regions
        on a flowcell during simulations

        :param read_sequences: Dict of raw sequences, accepting everything
        :param reads_decision: Dict of processed sequences after decisions
        :return:
        """
        # for control: all reads as they come out of the sequencer
        # total bases + (#reads * alpha)
        bases_total = np.sum([len(seq) for seq in read_sequences.values()])
        acquisition = self.batchsize * self.alpha
        self.time_control += (bases_total + acquisition)
        logging.info(f"time control: {self.time_control}")

        # for aeons: bases of the fully sequenced reads (accepted & unmapped) and of the truncated reads
        read_lengths_decision = np.array([len(seq) for seq in reads_decision.values()])
        n_reject = np.sum(np.where(read_lengths_decision == self.mu, 1, 0))
        bases_aeons = np.sum(read_lengths_decision)
        rejection_cost = n_reject * self.rho
        self.time_boss += (bases_aeons + acquisition + rejection_cost)
        logging.info(f"time boss-aeons: {self.time_boss}")



    def fill_cache(self, read_sequences: dict[str, str], reads_decision: dict[str, str]) -> None:
        """
        Write batches of simulated reads for convenient processing after the experiment
        Either only adds reads to a cache, or dumps them to file if it's time

        :param read_sequences: Dict of raw sequences
        :param reads_decision: Dict of sequences after BOSS decisions
        :return:
        """
        # helper function for both conditions
        def add_to_cache(seqs, cache):
            for rid, seq in seqs.items():
                cache[rid] = seq
        # add the current sequences to the cache
        add_to_cache(seqs=read_sequences, cache=self.cache_control)
        add_to_cache(seqs=reads_decision, cache=self.cache_boss)
        # check if time to dump and execute
        self._prep_dump(cond='control')
        self._prep_dump(cond='boss')



    def _prep_dump(self, cond: str) -> None:
        """
        Check if it's time to write cache to file, and execute if it is

        :param cond: String to determine which attributes to grab
        :return:
        """
        # grab the attributes of the condition
        curr_time = getattr(self, f'time_{cond}')
        dump_number = getattr(self, f'dump_n_{cond}')
        cache = getattr(self, f'cache_{cond}')
        # check if it's time to write out the next file
        if curr_time > (self.dumptime * dump_number):
            self._execute_dump(cond=cond, dump_number=dump_number, cache=cache)



    def _execute_dump(self, cond: str, dump_number: int, cache: dict[str, str]) -> None:
        """
        Write out the next cumulative batch file from the cache.
        This is for simulations only

        :param cond: String indicating which region on the flowcell
        :param dump_number: Running number of the dump to execute
        :param cache: Cache of read sequences
        :return:
        """
        logging.info(f'dump {cond} #{dump_number}. # of reads {len(list(cache.keys()))}')
        filename = f'00_reads/{cond}_{dump_number}.fa'
        # copy previous file to make cumulative
        previous_filename = f'00_reads/{cond}_{dump_number - 1}.fa'
        try:
            execute(f"cp {previous_filename} {filename}")
        except FileNotFoundError:
            # at the first batch, create empty 0th and copy to 1st file
            # to make sure we don't append to the same file multiple times
            # otherwise we have duplicate reads causing issues
            empty_file(previous_filename)
            execute(f"cp {previous_filename} {filename}")
        # writing operation
        with open(filename, "a") as f:
            for rid, seq in cache.items():
                r = random_id()
                fa_line = f'>{rid}.{r}\n{seq}\n'
                f.write(fa_line)
        # increment dump counter
        setattr(self, f'dump_n_{cond}', dump_number + 1)
        # reset cache
        setattr(self, f'cache_{cond}', dict())






